#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import ndarray
from .filter import Filter, BoxcarFilter


class Instrument:
    def __init__(self, name, filters, qes=None):
        self.name = name

        assert all([isinstance(f, Filter) for f in filters]), "All filters must be Filter instances."
        self.filters = filters
        self.pb_n = npb = len(filters)
        self.pb_names = [f.name for f in self.filters]

        if qes is not None:
            if isinstance(qes, (tuple, list, ndarray)):
                assert len(filters) == len(qes), "Number of QE profiles differs from the number of passbands."
                assert all([isinstance(qe, Filter) for qe in qes]), "All QE profiles must be Filter instances."
                self.qes = qes
            elif isinstance(qes, Filter):
                self.qes = npb * [qes]
        else:
            self.qes = npb * [BoxcarFilter('QE', 0., 10000.)]


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import atleast_2d, zeros, concatenate, ones, inf, full, squeeze, c_
from numba import njit, prange

from ...param import LParameter, NormalPrior as NP

@njit(parallel=True, fastmath=True)
def lbaseline(ltimes, lcids, pv, deg, cstart):
    pv = atleast_2d(pv)
    npv = pv.shape[0]
    npt = ltimes.size
    bl = zeros((npv, npt))
    for ipt in prange(npt):
        ilc = lcids[ipt]
        npl = deg[ilc] + 1
        leg = ones(npl)
        leg[1] = ltimes[ipt]
        for iln in range(npl):
            icoef = cstart[ilc] + iln
            for ipv in range(npv):
                bl[ipv, ipt] += pv[ipv, icoef] * leg[iln]
            if iln > 0 and npl > 1:
                leg[iln+1] = ((2*iln + 1)*ltimes[ipt]*leg[iln] - iln*leg[iln-1] ) / (iln+1)
    return bl


class LegendreBaseline:
    def __init__(self, nlegendre):
        self.nlegendre = nlegendre
        self.times = []
        self.timea = None
        self.lcslices = None
        self.nlc = 0
        self.ps = None

    def _init_p_baseline(self):
        """Baseline parameter initialisation.
        """

        self._baseline_times = [(t - t.mean()) / t.ptp() for t in self.times]
        self._baseline_timea = concatenate(self._baseline_times)

        if isinstance(self.nlegendre, int):
            self.nlegendre = full(self.nlc, self.nlegendre)

        assert self.nlegendre.size == len(self.times)

        bls = []
        for i, tn in enumerate(range(self.nlc)):
            fstd = self.fluxes[i].std()
            bls.append(LParameter(f'bli_{tn}', f'bl_intercept_{tn}', '', NP(1.0, fstd), bounds=(-inf, inf)))
            for ipoly in range(1, self.nlegendre[i] + 1):
                bls.append(
                    LParameter(f'bls_{tn}_{ipoly}', f'bl_c_{tn}_{ipoly}', '', NP(0.0, fstd), bounds=(-inf, inf)))
        self.ps.add_global_block('baseline', bls)
        self._sl_bl = self.ps.blocks[-1].slice
        self._start_bl = self.ps.blocks[-1].start
        self._bl_coeff_start = c_[[0], [self.nlegendre[:-1]+1]].cumsum()

    def baseline(self, pvp):
        """Multiplicative baseline"""
        pvp = atleast_2d(pvp)
        return squeeze(lbaseline(self._baseline_timea, self.lcids, pvp[:,self._sl_bl], self.nlegendre, self._bl_coeff_start))



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to model flux contamination in transit light curves.

Module to model flux contamination in transit light curves. The `pytransit.contamination` module contains two flux
contamination models: a physical model based on simulated stellar spectra (`SMContamination`), and a simple one
based on black body approximation (`BBContamination`). The models share an common interface.
"""

from os.path import join
from typing import Union, Iterable

import pandas as pd
import xarray as xa

from numba import njit
from matplotlib.pyplot import subplots, setp
from numpy import transpose, newaxis, uint32, ndarray, asarray, zeros_like, log, exp, ceil, linspace, array, nan
from pandas import DataFrame
from scipy.interpolate import interp1d, RegularGridInterpolator

from .instrument import Instrument
from ..stars import read_bt_settl_table, read_husser2013_table
from ..utils.phasecurves import planck


@njit
def contaminate_light_curve(flux: ndarray, contamination: ndarray, pbids: ndarray) -> ndarray:
    """Contaminates a transit light curve.

    Contaminates a transit light curve with npb passbands.

    Parameters
    ----------
    flux: 1d array-like
        Transit light curve with npb passbands.
    contamination: 1d array-like
        Array of per-passband contamination values.
    pbids: 1d array-like
        Passband indices that map each light curve element to a single passband.

    Returns
    -------
    Contaminated transit light curve
    """
    npt = flux.size
    contaminated_flux = zeros_like(flux)
    for i in range(npt):
        contaminated_flux[i] = contamination[pbids[i]] + (1.0-contamination[pbids[i]])*flux[i]
    return contaminated_flux


class _BaseContamination:
    """A base class to model flux contamination (blending) in transit light curves.

    A base class to model flux contamination (blending) in transit light curves.

    Notes
    -----
    This class does not implement all the necessary functionality, and is not meant to be used as-is.
    """

    def __init__(self, instrument: Instrument, ref_pb: str) -> None:
        """
        Parameters
        ----------
        instrument
            Instrument configuration
        ref_pb
            name of the reference passband
        """
        self.instrument = instrument
        self._ri = self.instrument.pb_names.index(ref_pb)
        self._rpb = ref_pb

    def relative_fluxes(self, teff:  Union[float, ndarray]):
        raise NotImplementedError

    def relative_flux_mixture(self, teffs: Iterable, fractions: Iterable):
        raise NotImplementedError

    def contamination(self, cr: float, teff1: float, teff2: float):
        raise NotImplementedError

    def exposure_times(self, teff, rtime, rflux=1.0, tflux=1.0):
        """Exposure times that give equal flux as in the reference passband

        Parameters
        ----------
        teff : float
            Effective stellar temperature [K]
        rtime : float
            Exposure time in the reference passband
        rflux : float, optional
            Flux in the reference passband with the reference exposure time
        tflux : float, optional
            Target flux in the reference passband

        """
        return rtime * tflux / rflux / self.relative_fluxes(teff)

    def c_as_pandas(self, ci, teff1, teff2):
        """Contamination as a pandas DataFrame."""
        return pd.DataFrame(self.contamination(ci, teff1, teff2),
                            columns=pd.Index(self.instrument.pb_names, name='passband'),
                            index=pd.Index(uint32(teff2), name='teff'))

    def c_as_xarray(self, ci, teff1, teff2):
        """Contamination as an xarray DataArray."""
        from xarray import DataArray
        return DataArray(self.contamination(ci, teff1, teff2),
                         name='contamination',
                         dims=['teff', 'passband'],
                         coords=[uint32(teff2), self.instrument.pb_names],
                         attrs={'TEff_1': teff1})

    def plot(self, teff1: float, teff2: float, cref, wlref=600, wlmin=305, wlmax=995, nwl=500, figsize=None, axs=None):
        """Plots the contamination model.

        Parameters
        ----------
        teff1
        teff2
        cref
        wlref
        wlmin
        wlmax
        nwl
        figsize

        Returns
        -------

        """
        wl = linspace(wlmin, wlmax, nwl)
        ft = (1 - cref) * self.relative_flux(teff1, wl, wlref)
        fc = cref * self.relative_flux(teff2, wl, wlref)

        if axs is None:
            fig, axs = subplots(2, 1, figsize=figsize, sharex='all', constrained_layout=True)
        else:
            fig = None

        axs[0].plot(wl, ft, label='host')
        axs[0].plot(wl, fc, label='contaminant')
        axs[1].plot(wl, fc / (ft + fc), 'k')

        for ax in axs:
            ax.axhline(cref, ls='--', c='0.75')
            ax.axvline(wlref, ls='--', c='0.75')

        axs[0].legend(fontsize='small')
        setp(axs, xlim=(wlmin, wlmax))
        setp(axs[0], ylabel='Relative flux')
        setp(axs[1], ylim=(-0.05, 1.05), xlabel='Wavelength [nm]', ylabel='Contamination')
        #fig.tight_layout()
        return fig


class BBContamination(_BaseContamination):
    """Third light contamination based on black body approximation.

    This class offers a simple black-body model for flux contamination in which the target star and the contaminant(s)
    are approximated as black bodies with effective temperatures Tt, Tc1, Tc2, ..., Tcn.
    """

    def __init__(self, instrument: Instrument, ref_pb: str, delta_l: float = 10):
        """

        Parameters
        ----------
        instrument
            Instrument configuration.
        ref_pb
            Reference passband name.
        delta_l
            Wavelength grid spacing in nm.
        """
        super().__init__(instrument, ref_pb)
        self._delta_l = delta_l
        self._wl_grids = []
        for f in self.instrument.filters:
            nwl = int(ceil((f.wl_max - f.wl_min) / self._delta_l))
            self._wl_grids.append(linspace(f.wl_min, f.wl_max, nwl))

    @staticmethod
    def absolute_flux(teff: float, wl: Union[float, Iterable]) -> ndarray:
        """The absolute flux given an effective temperature and wavelength.

        Parameters
        ----------
        teff
            The effective temperature in K
        wl
            The wavelength (or an array of) in nm

        Returns
        -------
        Black body spectral radiance.
        """
        return planck(teff, 1e-9 * wl)

    @staticmethod
    def relative_flux(teff: float, wl: Union[float, ndarray], wlref: float) -> ndarray:
        """The black body flux normalized to a given reference wavelength.

        Parameters
        ----------
        teff
            The effective temperature of the radiating body [K]
        wl
            The wavelength [nm]
        wlref
            The reference wavelength [nm]

        Returns
        -------
        The black body flux normalized to a given reference wavelength
        """
        return planck(teff, 1e-9 * wl) / planck(teff, 1e-9 * wlref)

    def absolute_fluxes(self, teff: float) -> ndarray:
        """Calculates the integrated absolute fluxes for all filters for a star with the given effective temperature

        Parameters
        ----------
        teff
            The effective temperature of the radiating body [K]

        Returns
        -------
        The integrated absolute fluxes for the filters in the instrument.
        """
        return array(
            [(self.absolute_flux(teff, g) * f(g)).mean() for f, g in zip(self.instrument.filters, self._wl_grids)])

    def relative_fluxes(self, teff: Union[float, ndarray]) -> ndarray:
        """Calculates the integrated fluxes for all filters normalized to the reference passband.

        Parameters
        ----------
        teff
            The effective temperature of the radiating body [K]

        Returns
        -------
            The integrated fluxes for all filters normalized to the reference passband.
        """
        fluxes = self.absolute_fluxes(teff)
        return fluxes / fluxes[self._ri]

    def contamination(self, cref: float, teff1: float, teff2: float) -> ndarray:
        """Calculates the contamination factors for all the filters given the contamination in the reference passband.

        Parameters
        ----------
        cref
            Reference passband contamination
        teff1
            Host star effective temperature
        teff2
            Contaminant effective temperature

        Returns
        -------
            Contamination factors for all the filters.
        """
        ft = (1.0 - cref) * self.relative_fluxes(teff1)
        fc = cref * self.relative_fluxes(teff2)
        return fc / (ft + fc)


class SMContamination(_BaseContamination):
    """A class that models flux contamination based on stellar spectrum models.
    """

    def __init__(self, instrument: Instrument, ref_pb: str = None, data: str = 'BT-SETTL') -> None:
        """

        Parameters
        ----------
        instrument
            Instrument configuration
        ref_pb
            name of the reference passband
        """
        super().__init__(instrument, ref_pb)

        self.data = data

        if data.lower() == 'bt-settl':
            self._spectra: DataFrame = read_bt_settl_table()
        elif data.lower() == 'husser2013':
            self._spectra: DataFrame = read_husser2013_table()
        else:
            raise ValueError('The dataset has to be either "BT-SETTL" or "Husser2013"')

        self.teff = teff = self._spectra.index.values
        self.wl = wl = self._spectra.columns.values

        # 2D interpolator for the spectrum table
        # --------------------------------------
        self._rgi = RegularGridInterpolator((self.teff, self.wl), self._spectra.values)

        # Dataframe indices
        # -----------------
        self.ipb = pd.Index(self.instrument.pb_names, name='passband')
        self.iteff = pd.Index(teff, name='teff')

        # Per-passband fluxes
        # -------------------
        self._compute_relative_flux_tables(0, ref_pb)

    def _integrate_fluxes(self, rdc=None):
        i, sp, wl = self.instrument, self._spectra, self.wl
        int_fluxes = transpose([(sp.values * f(wl) * qe(wl)).sum(1) for f, qe in zip(i.filters, i.qes)])
        self._af = pd.DataFrame(int_fluxes, columns=self.ipb, index=self.iteff)

    def _compute_relative_flux_tables(self, rdc=None, rpb=None):
        self._integrate_fluxes(rdc)
        af = self._af.values
        if rpb is not None:
            self._rpb = rpb
            self._ri = self.instrument.pb_names.index(rpb)
        self._rf = pd.DataFrame(af / af[:, self._ri][:, newaxis], columns=self.ipb, index=self.iteff)
        self._ip = interp1d(self.iteff.values, self._rf.values.T, bounds_error=False, fill_value=nan)
        self._ip_abs = interp1d(self.iteff.values, self._af.values.T, bounds_error=False, fill_value=nan)

    def absolute_flux(self, teff: float, wl: Union[float, Iterable]) -> ndarray:
        """The absolute flux given an effective temperature and a set of wavelength.

        Parameters
        ----------
        teff: float
            The effective temperature [K]
        wl: array-like
            The wavelengths to calculate the flux in [nm]

        Returns
        -------
        Spectral radiance.
        """
        return self._rgi((teff, wl))

    def relative_flux(self, teff: float, wl: Union[float, ndarray], wlref: float) -> ndarray:
        """The stellar flux normalized to a given reference wavelength.

        Parameters
        ----------
        teff: float
            The effective temperature of the radiating body [K]
        wl: array-like
            The wavelengths to calculate the flux in [nm]
        wlref: float
            The reference wavelength [nm]

        Returns
        -------
        The flux normalized to a given reference wavelength
        """
        return self.absolute_flux(teff, wl) / self.absolute_flux(teff, wlref)

    def relative_fluxes(self, teff: Union[float, ndarray], rdc=None, rpb=None):
        if (rpb is not None and rpb != self._rpb) or (rdc is not None):
            self._compute_relative_flux_tables(rdc, rpb)
        return self._ip(teff)

    def relative_flux_mixture(self, teffs, fractions, rdc=None):
        teffs = asarray(teffs)
        fractions = asarray(fractions)
        x = zeros_like(teffs)

        assert len(teffs) == len(fractions) + 1
        assert 0.0 < fractions.sum() < 1.0

        x[1:] = fractions
        x[0] = 1. - x[1:].sum()
        return (self.relative_fluxes(teffs, rdc) * x).sum(1)

    def contamination(self, cref: Union[float, ndarray], teff1: Union[float, ndarray], teff2: Union[float, ndarray], absolute: bool = False):
        """Contamination given reference contamination, host TEff, and contaminant TEff(s)

        Per-passband contamination given the contamination in the reference passband and TEffs of the two stars.

        Parameters
        ----------
        cref : float
            contamination in the reference passband
        teff1
            Effective stellar temperature [K]
        teff2
            Effective stellar temperature [K]

        Returns
        -------
        Per-passband contamination
        """
        if absolute:
            teff1, teff2 = asarray(teff1), asarray(teff2)
            a, b = self._ip_abs(teff1), self._ip_abs(teff2)
        else:
            cref, teff1, teff2 = asarray(cref), asarray(teff1), asarray(teff2)
            a = (1.0 - cref) * self._ip(teff1)
            b = cref * self._ip(teff2)
        return (b / (a + b)).T



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
from typing import Iterable

from astropy.table import Column
from matplotlib.pyplot import subplots, setp
from numpy import ndarray, array, median, floor, percentile, squeeze
from numpy.random import permutation

from .lpf import BaseLPF
from .. import TransitModel


class TransitLPF(BaseLPF):
    def __init__(self, name: str, passbands: Iterable, times: Iterable, fluxes: Iterable, errors: list = None,
                 pbids: list = None, covariates: list = None, wnids: list = None, tm: TransitModel = None,
                 nsamples: tuple = 1, exptimes: tuple = 0., init_data=True, result_dir: Path = None, tref: float = 0.0,
                 lnlikelihood: str = 'wn'):

        def transform_input(a):
            if isinstance(a, (list, tuple)):
                return a
            elif isinstance(a, Column):
                return [a.data.astype('d')]
            elif isinstance(a, ndarray):
                return [a]

        times = transform_input(times)
        fluxes = transform_input(fluxes)

        super().__init__(name, passbands, times, fluxes, errors, pbids, covariates, wnids, tm,
                         nsamples, exptimes, init_data, result_dir, tref, lnlikelihood)

    def plot_light_curve(self, model: str = 'de', figsize: tuple = (13, 4)):
        fig, ax = subplots(figsize=figsize, constrained_layout=True)
        tref = floor(self.timea.min())

        if model == 'de':
            pv = self.de.minimum_location
            err = 10 ** pv[7]
        elif model == 'mc':
            fc = array(self.posterior_samples())
            pv = permutation(fc)[:300]
            err = 10 ** median(pv[:, 7], 0)

        ax.errorbar(self.timea - tref, self.ofluxa, err, fmt='.', c='C0', alpha=0.75)

        if model == 'de':
            ax.plot(self.timea - tref, squeeze(self.flux_model(pv)), c='C0')
        if model == 'mc':
            flux_pr = self.flux_model(fc[permutation(fc.shape[0])[:1000]])
            flux_pc = array(percentile(flux_pr, [50, 0.15, 99.85, 2.5, 97.5, 16, 84], 0))
            [ax.fill_between(self.timea - tref, *flux_pc[i:i + 2, :], alpha=0.2, facecolor='C0') for i in
             range(1, 6, 2)]
            ax.plot(self.timea - tref, flux_pc[0], c='C0')
        setp(ax, xlim=self.timea[[0, -1]] - tref, xlabel=f'Time - {tref:.0f} [BJD]', ylabel='Normalised flux')
        return fig, ax


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from matplotlib.pyplot import subplots, setp
from numpy import sqrt, array, inf, int, s_, percentile, median, mean, round, zeros, atleast_2d, ceil, poly1d, polyfit
from numpy.random.mtrand import permutation
from uncertainties import ufloat, UFloat

try:
    import seaborn as sb
    with_seaborn = True
except ImportError:
    with_seaborn = False

from ..param.parameter import GParameter, LParameter
from ..param.parameter import UniformPrior as UP, NormalPrior as NP
from ..orbits.orbits_py import as_from_rhop, i_from_ba
from .ocllpf import OCLBaseLPF


def plot_estimates(x, p, ax, bwidth=0.8):
    ax.bar(x, p[4, :] - p[3, :], bwidth, p[3, :], alpha=0.25, fc='b')
    ax.bar(x, p[2, :] - p[1, :], bwidth, p[1, :], alpha=0.25, fc='b')
    [ax.plot((xx - 0.47 * bwidth, xx + 0.47 * bwidth), (pp[[0, 0]]), 'k') for xx, pp in zip(x, p.T)]


class OCLTTVLPF(OCLBaseLPF):
    def __init__(self, target: str, zero_epoch: float, period: float, passbands: list,
                 times: list = None, fluxes: list = None, errors: list = None, pbids: list = None, wnids: list = None,
                 nsamples: list = None, exptimes: list = None, cl_ctx=None, cl_queue=None):

        self.zero_epoch = zero_epoch if isinstance(zero_epoch, UFloat) else ufloat(zero_epoch, 1e-5)
        self.period = period if isinstance(period, UFloat) else ufloat(period, 1e-7)
        self.epoch = round((array([t.mean() for t in times]) - self.zero_epoch.n) / self.period.n).astype(int)

        super().__init__(target, passbands, times, fluxes, errors, pbids, nsamples=nsamples, exptimes=exptimes,
                         wnids=wnids, cl_ctx=cl_ctx, cl_queue=cl_queue)


    def _init_p_orbit(self):
        """Orbit parameter initialisation for a TTV model.

        The orbit part of the parameter vector will be [rho b tc_0 tc_1 ... tc_n].

        """

        # Basic orbital parameters
        # ------------------------
        porbit = [GParameter('rho', 'stellar_density', 'g/cm^3', UP(0.1, 25.0), (0, inf)),
                  GParameter('b', 'impact_parameter', 'R_s', UP(0.0, 1.0), (0, 1))]

        # Transit centers
        # ---------------
        self.epoch = round((array([t.mean() for t in self.times]) - self.zero_epoch.n) / self.period.n).astype(int)

        for e in self.epoch:
            tc = self.zero_epoch + e*self.period
            porbit.append(GParameter(f'tc_{e:d}', f'transit_centre_{e:d}', 'd', NP(tc.n, tc.s), (-inf, inf)))

        self.ps.add_global_block('orbit', porbit)
        self._start_tc = 2
        self._sl_tc = s_[self._start_tc:self._start_tc + self.nlc]

    def optimize_times(self, window):
        times, fluxes, pbids = [], [], []
        tcp = self.ps[self._sl_tc]
        for i in range(self.nlc):
            tc = tcp[i].prior.mean
            mask = abs(self.times[i] - tc) < 0.5*window/24.
            times.append(self.times[i][mask])
            fluxes.append(self.fluxes[i][mask])
        self._init_data(times, fluxes, self.pbids)

    def transit_model(self, pvp, copy=False):
        pvp = atleast_2d(pvp)
        pvp_cl = zeros([pvp.shape[0], 6 + self.nlc], "f")  # k tc p a i e w
        uv = zeros([pvp.shape[0], 2], "f")
        tc_end = 1 + self.nlc
        pvp_cl[:, 0:1] = sqrt(pvp[:, self._pid_k2])  # Radius ratio
        pvp_cl[:, 1:tc_end] = pvp[:, self._sl_tc]    # Transit centre and orbital period
        pvp_cl[:, tc_end + 0] = self.period.n
        pvp_cl[:, tc_end + 1] = a = as_from_rhop(pvp[:, 0], self.period.n)
        pvp_cl[:, tc_end + 2] = i_from_ba(pvp[:, 1], a)
        a, b = sqrt(pvp[:, self._sl_ld][:, 0]), 2. * pvp[:, self._sl_ld][:, 1]
        uv[:, 0] = a * b
        uv[:, 1] = a * (1. - b)
        flux = self.tm.evaluate_pv_ttv(pvp_cl, uv, copy=copy)
        return flux.T if copy else None

    def posterior_period(self, burn: int = 0, thin: int = 1) -> float:
        df = self.posterior_samples(burn, thin, False)
        tccols = [c for c in df.columns if 'tc' in c]
        tcs = median(df[tccols], 0)
        return mean((tcs[1:] - tcs[0]) / (self.epoch[1:] - self.epoch[0]))

    def plot_ttvs(self, burn=0, thin=1, axs=None, figsize=None, bwidth=0.8, fmt='h', windows=None, sigma=inf, nsamples=1000):
        assert fmt in ('d', 'h', 'min')
        multiplier = {'d': 1, 'h': 24, 'min': 1440}
        ncol = 1 if windows is None else len(windows)
        fig, axs = (None, axs) if axs is not None else subplots(1, ncol, figsize=figsize, sharey=True)
        df = self.posterior_samples(burn, thin, derived_parameters=False)
        tccols = [c for c in df.columns if 'tc' in c]
        df = df[tccols]
        s = df.std()
        m = (s < median(s) + sigma * s.std()).values
        df = df.iloc[:, m]
        epochs = self.epoch[m]

        samples = []
        for tcs in permutation(df.values)[:nsamples]:
            samples.append(tcs - poly1d(polyfit(epochs, tcs, 1))(epochs))
        samples = array(samples)
        p = multiplier[fmt] * percentile(samples, [50, 16, 84, 0.5, 99.5], 0)
        setp(axs, ylabel='Transit center - linear prediction [{}]'.format(fmt), xlabel='Transit number')
        if windows is None:
            plot_estimates(epochs, p, axs, bwidth)
            if with_seaborn:
                sb.despine(ax=axs, offset=15)
        else:
            setp(axs[1:], ylabel='')
            for ax, w in zip(axs, windows):
                m = (epochs > w[0]) & (epochs < w[1])
                plot_estimates(epochs[m], p[:, m], ax, bwidth)
                setp(ax, xlim=w)
                if with_seaborn:
                    sb.despine(ax=ax, offset=15)
        if fig:
            fig.tight_layout()
        return axs

    def plot_transits(self, ncols=4, figsize=(13, 11), remove=(), ylim=None):
        nt = len(self.times)
        nrows = int(ceil(nt / ncols))
        fig, axs = subplots(nrows, ncols, figsize=figsize, sharey='all', gridspec_kw=dict(hspace=0.01, wspace=0.01))

        if self.de is not None:
            pv = self.de.minimum_location
            fmodel = self.flux_model(pv).ravel()
        else:
            pv = None
            fmodel = None

        for i in range(nt):
            ax = axs.flat[i]

            # The light curve itself
            # ----------------------
            lc = '0.75' if i in remove else 'k'
            ax.plot(self.times[i], self.fluxes[i], '.', c=lc)

            # The fitted model, if available
            # ------------------------------
            if fmodel is not None:
                m = self.lcids == i
                ax.plot(self.times[i], fmodel[m], 'w', lw=4)
                ax.plot(self.times[i], fmodel[m], 'k', lw=1)

            # Transit centre prior
            # --------------------
            p = self.ps[self._start_tc + i].prior
            [ax.axvspan(p.mean - s * p.std, p.mean + s * p.std, alpha=0.15) for s in (3, 2, 1)]

            if self.de is not None:
                ax.axvline(pv[self._start_tc + i], c='k', zorder=-2)

            # The transit index
            # -----------------
            ax.text(0.05, 0.95, self.epoch[i], ha='left', va='top', transform=ax.transAxes)
            ax.text(0.05, 0.05, "({})".format(i), ha='left', va='bottom', transform=ax.transAxes, size='small')
            ax.set_xlim(self.times[i][[0, -1]])

        if ylim is not None:
            setp(axs, ylim=ylim)

        for ax in axs.flat[nt:]:
            ax.set_visible(False)

        setp(axs, xticks=[])
        fig.tight_layout()



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import seaborn as sb
import pandas as pd
import xarray as xa

from pathlib import Path
from time import strftime
from typing import Union, Iterable

from scipy.optimize import minimize
from numpy import ndarray, atleast_2d, inf, isfinite, where, clip, diag, full
from numpy.random import multivariate_normal
from emcee import EnsembleSampler
from matplotlib.pyplot import subplots, setp
from tqdm.auto import tqdm

from pytransit.utils.de import DiffEvol
from pytransit.param import ParameterSet, UniformPrior as UP, NormalPrior as NP


class LogPosteriorFunction:
    _lpf_name = 'LogPosteriorFunction'

    def __init__(self, name: str, result_dir: Union[Path, str] = '.'):
        """The Log Posterior Function class.

        Parameters
        ----------
        name: str
            Name of the log posterior function instance.
        """
        self.name = name
        self.result_dir = Path(result_dir if result_dir is not None else '.')

        # Declare high-level objects
        # --------------------------
        self.ps = None  # Parametrisation
        self.de = None  # Differential evolution optimiser
        self.sampler = None  # MCMC sampler
        self._local_minimization = None

        # Initialise the additional lnprior list
        # --------------------------------------
        self._additional_log_priors = []

        self._old_de_fitness = None
        self._old_de_population = None

    def print_parameters(self, columns: int = 2):
        columns = max(1, columns)
        for i, p in enumerate(self.ps):
            print(p.__repr__(), end=('\n' if i % columns == columns - 1 else '\t'))

    def _init_parameters(self):
        self.ps = ParameterSet()
        self.ps.freeze()

    def create_pv_population(self, npop=50):
        return self.ps.sample_from_prior(npop)

    def set_prior(self, parameter, prior, *nargs) -> None:
        if isinstance(parameter, str):
            descriptions = self.ps.descriptions
            names = self.ps.names
            if parameter in descriptions:
                parameter = descriptions.index(parameter)
            elif parameter in names:
                parameter = names.index(parameter)
            else:
                params = ', '.join([f"{ln} ({sn})" for ln, sn in zip(self.ps.descriptions, self.ps.names)])
                raise ValueError(f'Parameter "{parameter}" not found from the parameter set: {params}')

        if isinstance(prior, str):
            if prior.lower() in ['n', 'np', 'normal']:
                prior = NP(nargs[0], nargs[1])
            elif prior.lower() in ['u', 'up', 'uniform']:
                prior = UP(nargs[0], nargs[1])
            else:
                raise ValueError(f'Unknown prior "{prior}". Allowed values are (N)ormal and (U)niform.')

        self.ps[parameter].prior = prior

    def lnprior(self, pv: ndarray) -> Union[Iterable, float]:
        """Log prior density for a 1D or 2D array of model parameters.

        Parameters
        ----------
        pv: ndarray
            Either a 1D parameter vector or a 2D parameter array.

        Returns
        -------
            Log prior density for the given parameter vector(s).
        """
        return self.ps.lnprior(pv) + self.additional_priors(pv)

    def additional_priors(self, pv):
        pv = atleast_2d(pv)
        return sum([f(pv) for f in self._additional_log_priors], 0)

    def lnlikelihood(self, pv):
        raise NotImplementedError

    def lnposterior(self, pv):
        lnp = self.lnprior(pv) + self.lnlikelihood(pv)
        return where(isfinite(lnp), lnp, -inf)

    def __call__(self, pv):
        return self.lnposterior(pv)

    def optimize_local(self, pv0=None, method='powell'):
        if pv0 is None:
            if self.de is not None:
                pv0 = self.de.minimum_location
            else:
                pv0 = self.ps.mean_pv
        res = minimize(lambda pv: -self.lnposterior(pv), pv0, method=method)
        self._local_minimization = res

    def optimize_global(self, niter=200, npop=50, population=None, pool=None, lnpost=None, vectorize=True,
                        label='Global optimisation', leave=False, plot_convergence: bool = True, use_tqdm: bool = True,
                        plot_parameters: tuple = (0, 2, 3, 4)):

        lnpost = lnpost or self.lnposterior
        if self.de is None:
            self.de = DiffEvol(lnpost, clip(self.ps.bounds, -1, 1), npop, maximize=True, vectorize=vectorize, pool=pool)
            if population is None:
                self.de._population[:, :] = self.create_pv_population(npop)
            else:
                self.de._population[:, :] = population
        for _ in tqdm(self.de(niter), total=niter, desc=label, leave=leave, disable=(not use_tqdm)):
            pass

        if plot_convergence:
            fig, axs = subplots(1, 1 + len(plot_parameters), figsize=(13, 2), constrained_layout=True)
            rfit = self.de._fitness
            mfit = isfinite(rfit)

            if self._old_de_fitness is not None:
                m = isfinite(self._old_de_fitness)
                axs[0].hist(-self._old_de_fitness[m], facecolor='midnightblue', bins=25, alpha=0.25)
            axs[0].hist(-rfit[mfit], facecolor='midnightblue', bins=25)

            for i, ax in zip(plot_parameters, axs[1:]):
                if self._old_de_fitness is not None:
                    m = isfinite(self._old_de_fitness)
                    ax.plot(self._old_de_population[m, i], -self._old_de_fitness[m], 'kx', alpha=0.25)
                ax.plot(self.de.population[mfit, i], -rfit[mfit], 'k.')
                ax.set_xlabel(self.ps.descriptions[i])
            setp(axs, yticks=[])
            setp(axs[1], ylabel='Log posterior')
            setp(axs[0], xlabel='Log posterior')
            sb.despine(fig, offset=5)
        self._old_de_population = self.de.population.copy()
        self._old_de_fitness = self.de._fitness.copy()

    def sample_mcmc(self, niter: int = 500, thin: int = 5, repeats: int = 1, npop: int = None, population=None,
                    label='MCMC sampling', reset=True, leave=True, save=False, use_tqdm: bool = True, pool=None,
                    lnpost=None, vectorize: bool = True):

        if save and self.result_dir is None:
            raise ValueError('The MCMC sampler is set to save the results, but the result directory is not set.')

        lnpost = lnpost or self.lnposterior
        if self.sampler is None:
            if population is not None:
                pop0 = population
            elif hasattr(self, '_local_minimization') and self._local_minimization is not None:
                pop0 = multivariate_normal(self._local_minimization.x, diag(full(len(self.ps), 0.001 ** 2)), size=npop)
            elif self.de is not None:
                pop0 = self.de.population.copy()
            else:
                raise ValueError('Sample MCMC needs an initial population.')
            self.sampler = EnsembleSampler(pop0.shape[0], pop0.shape[1], lnpost, vectorize=vectorize, pool=pool)
        else:
            pop0 = self.sampler.chain[:, -1, :].copy()

        for i in tqdm(range(repeats), desc=label, disable=(not use_tqdm), leave=leave):
            if reset or i > 0:
                self.sampler.reset()
            for _ in tqdm(self.sampler.sample(pop0, iterations=niter, thin=thin), total=niter,
                          desc='Run {:d}/{:d}'.format(i + 1, repeats), leave=False, disable=(not use_tqdm)):
                pass
            if save:
                self.save(self.result_dir)
            pop0 = self.sampler.chain[:, -1, :].copy()

    def posterior_samples(self, burn: int = 0, thin: int = 1):
        fc = self.sampler.chain[:, burn::thin, :].reshape([-1, len(self.ps)])
        df = pd.DataFrame(fc, columns=self.ps.names)
        return df

    def plot_mcmc_chains(self, pid: int = 0, alpha: float = 0.1, thin: int = 1, ax=None):
        fig, ax = (None, ax) if ax is not None else subplots()
        ax.plot(self.sampler.chain[:, ::thin, pid].T, 'k', alpha=alpha)
        fig.tight_layout()
        return fig

    def save(self, save_path: Path = '.'):
        save_path = Path(save_path)
        npar = len(self.ps)

        if self.de:
            de = xa.DataArray(self.de.population, dims='pvector parameter'.split(), coords={'parameter': self.ps.names})
        else:
            de = None

        if self.sampler is not None:
            mc = xa.DataArray(self.sampler.chain, dims='pvector step parameter'.split(),
                              coords={'parameter': self.ps.names}, attrs={'ndim': npar, 'npop': self.sampler.nwalkers})
        else:
            mc = None

        ds = xa.Dataset(data_vars={'de_population': de, 'mcmc_samples': mc},
                        attrs={'created': strftime('%Y-%m-%d %H:%M:%S'), 'name': self.name})
        ds.to_netcdf(save_path.joinpath(f'{self.name}.nc'))

    def __repr__(self):
        return f"Target: {self.name}\nLPF: {self._lpf_name}"



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.s
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

import pandas as pd
from matplotlib.pyplot import subplots, setp
from numpy import arange, concatenate, zeros, inf, atleast_2d, where, repeat, squeeze, argsort, isfinite, ndarray, \
    floor, ceil, percentile, median
from numpy.random import uniform, permutation

from pytransit import LinearModelBaseline
from pytransit.contamination import SMContamination, Instrument
from pytransit.contamination.filter import sdss_g, sdss_r, sdss_i, sdss_z
from pytransit.lpf.cntlpf import PhysContLPF, map_pv_pclpf, map_ldc, contaminate
from pytransit.lpf.tesslpf import downsample_time
from pytransit.orbits import epoch
from pytransit.param.parameter import NormalPrior as NP, UniformPrior as UP, GParameter, PParameter
from pytransit.utils.misc import fold


class BaseTGCLPF(PhysContLPF):
    """Log posterior function that combined TESS light curve with ground-based transit light curves.

    Example log posterior function to model a TESS light curve jointly with with ground-based light curves observed in
    arbitrary passbands. The TESS light curve is assumed to contain an unknown amount of light contamination, and the
    ground-based observations are assumed to all be contaminated by an identical source, but the contamination follows
    a physical passband-dependent model.

    **Note:** This LPF is meant to be inherited by an LPF that implements the `read_data` method.
    """

    def __init__(self, name: str, use_ldtk: bool = False, tm = None):
        self.result_dir = Path('.')
        self._stess = None
        self._ntess = None

        times, fluxes, pbnames, pbs, wns, covs = self.read_data()
        pbids = pd.Categorical(pbs, categories=pbnames).codes
        wnids = arange(len(times))
        tref = floor(concatenate(times).min())

        self.wns = wns
        PhysContLPF.__init__(self, name, passbands=pbnames, times=times, fluxes=fluxes, pbids=pbids, wnids=wnids,
                             covariates=covs, tref=tref, tm=tm)

    def read_data(self):
        """Read in the TESS light curve and the ground-based data.

        Read in the TESS light curve and the ground-based data. This method needs to be implemented by the class inheriting
        the base ´TESSGBLPF`.
        """
        raise NotImplementedError

    def _init_p_planet(self):
        ps = self.ps
        pk2 = [PParameter('k2_app', 'apparent_area_ratio', 'A_s', UP(0.01**2, 0.30**2), (0., inf))]
        pcn = [GParameter('k2_true', 'true_area_ratio', 'As', UP(0.01**2, 0.75**2), bounds=(1e-8, inf)),
               GParameter('teff_h', 'host_teff', 'K', UP(2500, 12000), bounds=(2500, 12000)),
               GParameter('teff_c', 'contaminant_teff', 'K', UP(2500, 12000), bounds=(2500, 12000)),
               GParameter('k2_app_tess', 'tess_apparent_area_ratio', 'A_s', UP(0.01**2, 0.30**2), (0., inf))]
        ps.add_passband_block('k2', 1, 1, pk2)
        self._pid_k2 = repeat(ps.blocks[-1].start, self.npb)
        self._start_k2 = ps.blocks[-1].start
        self._sl_k2 = ps.blocks[-1].slice
        ps.add_global_block('contamination', pcn)
        self._pid_cn = arange(ps.blocks[-1].start, ps.blocks[-1].stop)
        self._sl_cn = ps.blocks[-1].slice

    def _init_baseline(self):
        self._add_baseline_model(LinearModelBaseline(self))

    def create_pv_population(self, npop: int = 50) -> ndarray:
        pvp = zeros((0, len(self.ps)))
        npv, i = 0, 0
        while npv < npop and i < 10:
            pvp_trial = self.ps.sample_from_prior(npop)
            pvp_trial[:, 5] = pvp_trial[:, 4]
            cref = uniform(0, 0.99, size=npop)
            pvp_trial[:, 5] = pvp_trial[:, 4] / (1. - cref)
            lnl = self.lnposterior(pvp_trial)
            ids = where(isfinite(lnl))
            pvp = concatenate([pvp, pvp_trial[ids]])
            npv = pvp.shape[0]
            i += 1
        pvp = pvp[:npop]
        return pvp

    def posterior_samples(self, burn: int = 0, thin: int = 1, derived_parameters: bool = True):
        df = super().posterior_samples(burn, thin, derived_parameters)
        if derived_parameters:
            df['ctess'] = 1 - df.k2_app_tess / df.k2_true
        return df

    def _init_instrument(self):
        """Set up the instrument and contamination model."""
        self.instrument = Instrument('example', [sdss_g, sdss_r, sdss_i, sdss_z])
        self.cm = SMContamination(self.instrument, "i'")
        self.add_prior(lambda pv: where(pv[:, 4] < pv[:, 5], 0, -inf))
        self.add_prior(lambda pv: where(pv[:, 8] < pv[:, 5], 0, -inf))

    def transit_model(self, pvp):
        pvp = atleast_2d(pvp)
        cnt = zeros((pvp.shape[0], self.npb))
        pvt = map_pv_pclpf(pvp)
        pvt[:,1] -= self._tref
        ldc = map_ldc(pvp[:, self._sl_ld])
        flux = self.tm.evaluate_pv(pvt, ldc)
        cnt[:, 0] = 1 - pvp[:, 8] / pvp[:, 5]
        cnref = 1. - pvp[:, 4] / pvp[:, 5]
        cnt[:, 1:] = self.cm.contamination(cnref, pvp[:, 6], pvp[:, 7])
        return contaminate(flux, cnt, self.lcids, self.pbids)

    def plot_folded_tess_transit(self, solution: str = 'de', pv: ndarray = None, binwidth: float = 1,
                                 plot_model: bool = True, plot_unbinned: bool = True, plot_binned: bool = True,
                                 xlim: tuple = None, ylim: tuple = None, ax=None, figsize: tuple = None):

        if pv is None:
            if solution.lower() == 'local':
                pv = self._local_minimization.x
            elif solution.lower() in ('de', 'global'):
                pv = self.de.minimum_location
            elif solution.lower() in ('mcmc', 'mc'):
                pv = self.posterior_samples().median().values
            else:
                raise NotImplementedError("'solution' should be either 'local', 'global', or 'mcmc'")

        if ax is None:
            fig, ax = subplots(figsize=figsize)
        else:
            fig, ax = None, ax

        ax.autoscale(enable=True, axis='x', tight=True)

        etess = self._ntess
        t = self.timea[:etess]
        fo = self.ofluxa[:etess]
        fm = squeeze(self.transit_model(pv))[:etess]
        bl = squeeze(self.baseline(pv))[:etess]

        phase = 24 * pv[1] * (fold(t, pv[1], pv[0], 0.5) - 0.5)
        sids = argsort(phase)
        phase = phase[sids]
        bp, bf, be = downsample_time(phase, (fo / bl)[sids], binwidth / 60)
        if plot_unbinned:
            ax.plot(phase, (fo / bl)[sids], 'k.', alpha=0.1, ms=2)
        if plot_binned:
            ax.errorbar(bp, bf, be, fmt='ko', ms=3)
        if plot_model:
            ax.plot(phase, fm[sids], 'k')
        setp(ax, ylim=ylim, xlim=xlim, xlabel='Time - T$_c$ [h]', ylabel='Normalised flux')

        if fig is not None:
            fig.tight_layout()
        return fig

    def plot_gb_transits(self, solution: str = 'de', pv: ndarray = None, figsize: tuple = None, axes=None, ncol: int = 4,
                         xlim: tuple = None, ylim: tuple = None, remove_baseline: bool = True, n_samples: int = 1500):

        solution = solution.lower()
        samples = None

        if pv is None:
            if solution == 'local':
                pv = self._local_minimization.x
            elif solution in ('de', 'global'):
                solution = 'global'
                pv = self.de.minimum_location
            elif solution in ('mcmc', 'mc'):
                solution = 'mcmc'
                samples = self.posterior_samples(derived_parameters=False)
                samples = permutation(samples.values)[:n_samples]
                pv = median(samples, 0)
            else:
                raise NotImplementedError("'solution' should be either 'local', 'global', or 'mcmc'")

        nlc = self.nlc - self._stess
        nrow = int(ceil(nlc / ncol))

        if axes is None:
            fig, axs = subplots(nrow, ncol, figsize=figsize, sharex='all', sharey='all', squeeze=False)
        else:
            fig, axs = None, axes

        [ax.autoscale(enable=True, axis='x', tight=True) for ax in axs.flat]

        if remove_baseline:
            if solution == 'mcmc':
                fbasel = median(self.baseline(samples), axis=0)
                fmodel, fmodm, fmodp = percentile(self.transit_model(samples), [50, 0.5, 99.5], axis=0)
            else:
                fbasel = squeeze(self.baseline(pv))
                fmodel, fmodm, fmodp = squeeze(self.transit_model(pv)), None, None
            fobs = self.ofluxa / fbasel
        else:
            if solution == 'mcmc':
                fbasel = median(self.baseline(samples), axis=0)
                fmodel, fmodm, fmodp = percentile(self.flux_model(samples), [50, 1, 99], axis=0)
            else:
                fbasel = squeeze(self.baseline(pv))
                fmodel, fmodm, fmodp = squeeze(self.flux_model(pv)), None, None
            fobs = self.ofluxa

        etess = self._stess
        t0, p = pv[[0, 1]]

        for i in range(nlc):
            ax = axs.flat[i]
            sl = self.lcslices[etess + i]
            t = self.times[etess + i]
            e = epoch(t.mean(), t0, p)
            tc = t0 + e * p
            tt = 24 * (t - tc)
            ax.plot(tt, fobs[sl], 'k.', alpha=0.2)
            ax.plot(tt, fmodel[sl], 'k')

            if solution == 'mcmc':
                ax.fill_between(tt, fmodm[sl], fmodp[sl], zorder=-100, alpha=0.2, fc='k')

            if not remove_baseline:
                ax.plot(tt, fbasel[sl], 'k--', alpha=0.2)

        setp(axs, xlim=xlim, ylim=ylim)
        setp(axs[-1, :], xlabel='Time - T$_c$ [h]')
        setp(axs[:, 0], ylabel='Normalised flux')

        if fig is not None:
            fig.tight_layout()
        return fig




#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Log posterior functions for transit modelling and parameter estimation.

A log posterior function (LPF) class creates a basis for Bayesian parameter estimation from transit light curves.
In PyTransit, LPFs are a bit more than what the name implies. An LPF stores the observations, model priors, etc.
It also contains methods for posterior optimisation and MCMC sampling.

"""

from .lpf import BaseLPF
from .phasecurvelpf import PhaseCurveLPF
from .baselines.legendrebaseline import LegendreBaseline


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Contaminated TESS LPF

This module contains the log posterior function (LPF) for a TESS light curve with possible third-light
contamination from an unresolved source inside the photometry aperture.
"""
import pandas as pd
from pathlib import Path
from typing import Union, Iterable, List

from astropy.table import Table
from matplotlib.pyplot import subplots, setp
from numba import njit
from astropy.io.fits import getval
from numpy import repeat, inf, where, newaxis, squeeze, atleast_2d, isfinite, concatenate, zeros, sqrt, zeros_like, \
    ones, median, any, arange, argsort
from numpy.random.mtrand import uniform
from pytransit import BaseLPF
from pytransit.lpf.tesslpf import TESSLPF, downsample_time
from pytransit.orbits import as_from_rhop, i_from_ba
from pytransit.param import UniformPrior as UP, NormalPrior as NP, PParameter, GParameter
from ldtk import tess
from pytransit.utils.misc import fold
from scipy.ndimage import label
from uncertainties import ufloat, UFloat


@njit(fastmath=True)
def map_pv(pv, i):
    pv = atleast_2d(pv)
    pp = pv[:,2+4*i:6+4*i]
    pvt = zeros((pv.shape[0], 7))
    pvt[:,0]   = sqrt(pp[:,3])
    pvt[:,1:3] = pp[:,0:2]
    pvt[:,  3] = as_from_rhop(pv[:,0], pp[:,1])
    pvt[:,  4] = i_from_ba(pp[:,2], pvt[:,3])
    return pvt


@njit(fastmath=True, cache=False)
def map_ldc(ldc):
    ldc = atleast_2d(ldc)
    uv = zeros_like(ldc)
    a, b = sqrt(ldc[:,0::2]), 2.*ldc[:,1::2]
    uv[:,0::2] = a * b
    uv[:,1::2] = a * (1. - b)
    return uv

class CMPTESSLPF(BaseLPF):
    """Contaminated TESS LPF for multiplanet systems

    This class implements a log posterior function for a TESS light curve that allows for unknown flux contamination.
    The amount of flux contamination is not constrained.
    """
    def __init__(self, name: str, fname: Union[Path, str], nplanets: int, zero_epochs: List, periods: List,
                 nsamples: int = 1, bldur: float = 0.1, trdur: float = 0.04, use_pdc: bool = False):

        self._fname = fname
        self.nplanets = nplanets

        for i,t0 in enumerate(zero_epochs):
            if isinstance(t0, float):
                zero_epochs[i] = ufloat(t0, 1e-3)

        for i, pr in enumerate(periods):
            if isinstance(pr, float):
                periods[i] = ufloat(pr, 1e-5)

        self.zero_epochs = zero_epochs
        self.periods = periods
        self.bjdrefi = getval(fname, 'bjdrefi', 1)
        self.bldur = bldur
        self.trdur = trdur

        times, fluxes, pbnames, pbs, wns, covs = self.read_data()
        pbids = pd.Categorical(pbs, categories=pbnames).codes
        wnids = arange(len(times))
        self.wns = wns

        BaseLPF.__init__(self, name, passbands=pbnames, times=times, fluxes=fluxes, pbids=pbids, wnids=wnids,
                         covariates=covs, nsamples=nsamples, exptimes=[0.00139])

    def read_data(self):
        raise NotImplementedError

    def _init_p_orbit(self):
        """Orbit parameter initialisation.
        """
        porbit = [GParameter('rho', 'stellar_density',  'g/cm^3', UP(0.1, 25.0), (0,    inf))]
        self.ps.add_global_block('orbit', porbit)

    def _init_p_planet(self):
        ps = self.ps
        pp = [GParameter('ctess', 'tess_contamination', '', UP(0.0, 0.99), (0, 1))]
        for i in range(1,self.nplanets+1):
            pp.extend([
                GParameter(f'tc_{i}', f'zero_epoch_{i}',       'd', NP(0.0, 0.1), (-inf, inf)),
                GParameter(f'p_{i}',  f'period_{i}',           'd', NP(1.0, 1e-5), (0, inf)),
                GParameter(f'b_{i}',  f'impact_parameter_{i}', 'R_s', UP(0.0, 1.0), (0, inf)),
                GParameter(f'k2_true_{i}', f'true_area_ratio_{i}', 'A_s', UP(0.02**2, 0.1**2), (0.02**2, 0.1**2))
            ])
        ps.add_global_block('planets', pp)
        self._start_pl = ps.blocks[-1].start
        self._sl_pl = ps.blocks[-1].slice

    def transit_model(self, pv, copy=True, planets=None):
        pv = atleast_2d(pv)
        flux = ones([pv.shape[0],self.timea.size])
        ldc = map_ldc(pv[:,self._sl_ld])
        planets = planets if planets is not None else arange(self.nplanets)
        for i in planets:
            pvp = map_pv(pv, i)
            flux += self.tm.evaluate_pv(pvp, ldc, copy) - 1.
        cnt = pv[:,1]
        return squeeze(cnt[:, newaxis] + (1. - cnt[:, newaxis]) * flux)

    def create_pv_population(self, npop=50):
        return self.ps.sample_from_prior(npop)

    def plot_folded_planets(self, bwidth: float = 10):
        fig, axs = subplots(1, self.nplanets, figsize=(13, 6), sharey=True)
        pv = self.de.minimum_location.copy()

        for ipl in range(self.nplanets):
            planets = set(arange(self.nplanets))
            planets.remove(ipl)
            mflux = self.transit_model(pv, planets=[ipl])
            rflux = self.transit_model(pv, planets=planets)
            oflux = self.ofluxa / rflux
            phase = (fold(self.timea, pv[3 + 4 * ipl], pv[2 + 4 * ipl], 0.5) - 0.5) * pv[3 + 4 * ipl]
            m = abs(phase) < 0.5 * self.bldur
            sids = argsort(phase[m])
            phase, mflux, oflux = phase[m][sids], mflux[m][sids], oflux[m][sids]
            bp, bf, be = downsample_time(phase, oflux, bwidth / 24 / 60)
            axs[ipl].errorbar(bp, bf, be, fmt='ok')
            axs[ipl].plot(phase, oflux, '.', alpha=0.25)
            axs[ipl].plot(phase, mflux, 'k')

        setp(axs, xlim=(-0.5 * self.bldur, 0.5 * self.bldur), ylim=(0.996, 1.004))
        fig.tight_layout()
        return fig


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from matplotlib.pyplot import subplots, setp
from numba import njit
from numpy import inf, repeat, atleast_2d, sqrt, arctan2, squeeze, ndarray, argsort, array, unique, arange
from seaborn import despine

from .loglikelihood.fmcloglikelihood import FrozenMultiCeleriteLogLikelihood
from .lpf import BaseLPF
from .. import UniformModel
from .tesslpf import downsample_time
from ..param import GParameter, PParameter, UniformPrior as UP, NormalPrior as NP
from ..orbits import as_from_rhop, i_from_ba, i_from_baew, eclipse_phase, d_from_pkaiews, epoch
from ..utils.misc import fold


@njit
def set_depths(f, pvp, pbids, lcids, ik2, sfr):
    f = atleast_2d(f)
    pvp = atleast_2d(pvp)
    npt = f.shape[1]
    for i in range(npt):
        ipb = pbids[lcids[i]]
        f[:, i] = 1 + (f[:, i] - 1.) * pvp[:, sfr + ipb]
    return f


class EclipseLPF(BaseLPF):

    def _init_p_limb_darkening(self):
        pass

    def _init_p_orbit(self):
        porbit = [
            GParameter('tc', 'zero epoch', 'd', NP(0.0, 0.1), (-inf, inf)),
            GParameter('p', 'period', 'd', NP(1.0, 1e-5), (0, inf)),
            GParameter('rho', 'stellar density', 'g/cm^3', UP(0.1, 25.0), (0, inf)),
            GParameter('b', 'impact parameter', 'R_s', UP(0.0, 1.0), (0, 1)),
            GParameter('secw', 'sqrt(e) cos(w)', '', UP(-0.5, 0.5), (-1, 1)),
            GParameter('sesw', 'sqrt(e) sin(w)', '', UP(-0.5, 0.5), (-1, 1))]
        self.ps.add_global_block('orbit', porbit)

    def _init_p_planet(self):
        """Planet parameter initialisation.
        """
        pk2 = [PParameter('k2', 'area ratio', 'A_s', UP(0.01 ** 2, 0.2 ** 2), (0, inf))]
        self.ps.add_global_block('k2', pk2)
        self._pid_k2 = repeat(self.ps.blocks[-1].start, 1)
        self._start_k2 = self.ps.blocks[-1].start
        self._sl_k2 = self.ps.blocks[-1].slice
        self._ix_k2 = self._start_k2

        pfr = [PParameter(f'fr_{pb}', 'Flux ratio', '', UP(0, 1), (0, 1)) for pb in self.passbands]
        self.ps.add_passband_block('fr', len(self.passbands), 1, pfr)
        self._pid_fr = repeat(self.ps.blocks[-1].start, self.npb)
        self._start_fr = self.ps.blocks[-1].start
        self._sl_fr = self.ps.blocks[-1].slice

    def _init_lnlikelihood(self):
        self._add_lnlikelihood_model(FrozenMultiCeleriteLogLikelihood(self))

    def _post_initialisation(self):
        super()._post_initialisation()
        self.tm = UniformModel(eclipse=True)
        self.tm.set_data(self.timea - self._tref, self.lcids, self.pbids, self.nsamples, self.exptimes)

    def transit_model(self, pv, copy=True):
        pv = atleast_2d(pv)
        zero_epoch = pv[:, 0] - self._tref
        period = pv[:, 1]
        smaxis = as_from_rhop(pv[:, 2], period)
        inclination = i_from_ba(pv[:, 3], smaxis)
        radius_ratio = sqrt(pv[:, 6:7])
        eccentricity = pv[:, 4] ** 2 + pv[:, 5] ** 2
        omega = arctan2(pv[:, 5], pv[:, 4])
        fmod = self.tm.evaluate(radius_ratio, zero_epoch, period, smaxis, inclination, eccentricity, omega)
        fmod = set_depths(fmod, pv, self.pbids, self.lcids, self._start_k2, self._start_fr)
        return squeeze(fmod)

    def create_pv_population(self, npop):
        return self.ps.sample_from_prior(npop)

    def plot_light_curves(self, pv=None, figsize=None, remove_baseline: bool = False):
        if pv is None:
            if self.de is not None:
                pv = self.de.minimum_location
            else:
                pv = self.ps.mean_pv

        tc, p, rho, b, secw, sesw, k2 = pv[:7]
        a = as_from_rhop(rho, p)
        e = secw ** 2 + sesw ** 2
        w = arctan2(sesw, secw)
        i = i_from_baew(b, a, e, w)
        ec = tc + eclipse_phase(p, i, e, w)
        t14 = d_from_pkaiews(p, sqrt(k2), a, i, e, w, -1)
        eps = array([epoch(t.mean(), ec, p) for t in self.times])
        uep = unique(eps)
        nep = uep.size
        npb = self.npb

        fig, axs = subplots(nep, npb, sharey='all', sharex='all', figsize=figsize)
        emap = {e: ied for e, ied in zip(uep, arange(nep))}
        fmodel = self.flux_model(self.de.minimum_location)
        bline = self._lnlikelihood_models[0].predict_baseline(pv)

        for ilc in range(self.nlc):
            iep = emap[eps[ilc]]
            ipb = self.pbids[ilc]
            ax = axs[iep, ipb]
            time = 24 * (self.times[ilc] - (ec + eps[ilc] * p))
            ax.plot(time, self.fluxes[ilc])
            if remove_baseline:
                ax.plot(time, fmodel[self.lcslices[ilc]], 'k')
            else:
                ax.plot(time, fmodel[self.lcslices[ilc]] + bline[self.lcslices[ilc]] - 1, 'k')

            ax.axvspan(-24 * 0.5 * t14, 24 * 0.5 * t14, alpha=0.25)
        setp(axs[-1], xlabel='Time - T$_c$ [h]')
        setp(axs[:, 0], ylabel='Normalised flux')
        fig.tight_layout()
        return fig

    def plot_folded_transit(self, solution: str = 'de', pv: ndarray = None, binwidth: float = 1,
                            plot_model: bool = True, plot_unbinned: bool = True, plot_binned: bool = True,
                            xlim: tuple = None, ylim: tuple = None, ax=None, figsize: tuple = None, malpha=0.1):

        # TODO: Doesn't take the passband information into account yet -> FIX!

        if pv is None:
            if solution.lower() == 'local':
                pv = self._local_minimization.x
            elif solution.lower() in ('de', 'global'):
                pv = self.de.minimum_location
            elif solution.lower() in ('mcmc', 'mc'):
                pv = self.posterior_samples(derived_parameters=False).median().values
            else:
                raise NotImplementedError("'solution' should be either 'local', 'global', or 'mcmc'")

        if ax is None:
            fig, ax = subplots(figsize=figsize)
        else:
            fig, ax = None, ax

        ax.autoscale(enable=True, axis='x', tight=True)

        t = self.timea
        fo = self.ofluxa
        fm = squeeze(self.transit_model(pv))
        bl = squeeze(self.baseline(pv))

        phase = 24 * pv[1] * (fold(t, pv[1], pv[0], 0.0) - 0.5)
        sids = argsort(phase)
        phase = phase[sids]
        if plot_unbinned:
            ax.plot(phase, (fo / bl)[sids], 'k.', alpha=malpha, ms=2)
        if plot_binned:
            bp, bf, be = downsample_time(phase, (fo / bl)[sids], binwidth / 60)
            ax.errorbar(bp, bf, be, fmt='ko', ms=3)
        if plot_model:
            ax.plot(phase, fm[sids], 'k')
        setp(ax, ylim=ylim, xlim=xlim, xlabel='Time - T$_c$ [h]', ylabel='Normalised flux')

        if fig is not None:
            fig.tight_layout()
        return fig

    def plot_flux_ratio_posteriors(self, figsize=None):
        df = self.posterior_samples(derived_parameters=False)
        frcols = [c for c in df.columns if 'fr' in c]
        fig, axs = subplots(1, len(frcols), sharex='all', sharey='all', figsize=figsize)
        for i, c in enumerate(frcols):
            axs[i].hist(df[c])
            axs[i].set_xlabel(f"{c.strip('_s').split('_')[-1]} flux ratio")
        setp(axs, yticks=[])
        despine(fig, offset=10)
        fig.tight_layout()
        return fig


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, Optional

from numba import njit
from numpy import ndarray, squeeze, zeros, asarray, ones
from .numba.ma_uniform_nb import uniform_model_v, uniform_model_s, uniform_model_pv
from .transitmodel import TransitModel

__all__ = ['EclipseModel']

npfloat = Union[float, ndarray]


class EclipseModel(TransitModel):

    def __init__(self) -> None:
        super().__init__()

    def evaluate(self, k: npfloat, t0: npfloat, p: npfloat, a: npfloat, i: npfloat, e: npfloat = None, w: npfloat = None,
                 fr: npfloat = None, multiplicative: bool = False, copy: bool = True) -> ndarray:
        """Evaluates a secondary eclipse model for a set of scalar or vector parameters.

        Parameters
        ----------
        k
            Radius ratio(s) either as a single float, 1D vector, or 2D array.
        t0
            Transit center(s) as a float or a 1D vector.
        p
            Orbital period(s) as a float or a 1D vector.
        a
            Orbital semi-major axis (axes) divided by the stellar radius as a float or a 1D vector.
        i
            Orbital inclination(s) as a float or a 1D vector.
        e
            Orbital eccentricity as a float or a 1D vector.
        w
            Argument of periastron as a float or a 1D vector.
        fr
            Planet-star flux ratio as a float or a 1D vector.
        multiplicative
            If True, will return the fraction of the visible planet disk area to the total planet disk area
        copy

        Notes
        -----
        The model can be evaluated either for one set of parameters or for many sets of parameters simultaneously.
        The orbital parameters can be given either as a float or a 1D array-like (preferably ndarray for optimal speed.)

        Returns
        -------
        Transit model
        """

        k = asarray(k)

        # Scalar parameters branch
        # ------------------------
        if isinstance(t0, float):
            e = 0. if e is None else e
            w = 0. if w is None else w

            flux = uniform_model_s(self.time, k, t0, p, a, i, e, w, self.lcids, self.pbids, self.nsamples,
                                   self.exptimes, zsign=-1.0)

        # Parameter population branch
        # ---------------------------
        else:
            npv = t0.size
            e = zeros(npv) if e is None else e
            w = zeros(npv) if w is None else w

            if k.ndim == 1:
                k = k.reshape((k.size,1))

            flux = uniform_model_v(self.time, k, t0, p, a, i, e, w, self.lcids, self.pbids, self.nsamples, self.exptimes,
                                   zsign=-1.0)

        if fr is not None:
            flux = 1.0 + (flux - 1.0) * fr
        elif multiplicative:
            flux = 1.0 + (flux - 1.0) / k**2

        return squeeze(flux)

    def evaluate_ps(self, k: float, t0: float, p: float, a: float, i: float, e: float = 0., w: float = 0., fr: Optional[float] = None) -> ndarray:
        """Evaluate the transit model for a set of scalar parameters.

         Parameters
         ----------
         k : array-like
             Radius ratio(s) either as a single float or an 1D array.
         t0 : float
             Transit center as a float.
         p : float
             Orbital period as a float.
         a : float
             Orbital semi-major axis divided by the stellar radius as a float.
         i : float
             Orbital inclination(s) as a float.
         e : float, optional
             Orbital eccentricity as a float.
         w : float, optional
             Argument of periastron as a float.
         fr : float, optional
             Planet-star flux ratio as a float.


         Notes
         -----
         This version of the `evaluate` method is optimized for calculating a single transit model (such as when using a
         local optimizer). If you want to evaluate the model for a large number of parameters simultaneously, use either
         `evaluate` or `evaluate_pv`.

         Returns
         -------
         ndarray
             Modelled flux as a 1D ndarray.
         """
        if self.time is None:
            raise ValueError("Need to set the data before calling the transit model.")

        k = asarray(k)
        flux = uniform_model_s(self.time, k, t0, p, a, i, e, w, self.lcids, self.pbids, self.nsamples, self.exptimes, zsign=-1.0)
        if fr is not None:
            flux = 1.0 + (flux - 1.0) * fr
        return squeeze(flux)

    def evaluate_pv(self, pvp: ndarray, fr = None) -> ndarray:
        """Evaluate the transit model for a 2D parameter array.

         Parameters
         ----------
         pvp
             Parameter array with a shape `(npv, npar)` where `npv` is the number of parameter vectors, and each row
             contains a set of parameters `[k, t0, p, a, i, e, w]`. The radius ratios can also be given per passband,
             in which case the row should be structured as `[k_0, k_1, k_2, ..., k_npb, t0, p, a, i, e, w]`.

         Notes
         -----
         This version of the `evaluate` method is optimized for calculating several models in parallel, such as when
         using *emcee* for MCMC sampling.

         Returns
         -------
         ndarray
             Modelled flux either as a 1D or 2D ndarray.
         """
        assert self.time is not None, "Need to set the data before calling the transit model."
        flux = uniform_model_pv(self.time, pvp, self.lcids, self.pbids, self.nsamples, self.exptimes, zsign=-1.0)
        if fr is not None:
            flux = 1.0 + (flux - 1.0) * fr
        return squeeze(flux)



# -*- coding: utf-8 -*-

#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""QPower2 transit model.

This module implements the qpower2 transit model by Maxted & Gill (A&A, 622, A33, 2019).
"""

from numpy import any, sqrt, pi, arccos, ones_like, atleast_2d, zeros, atleast_1d, nan, copysign, fmax, floor
from numba import njit, prange

from ...orbits.taylor_z import vajs_from_paiew, z_taylor_st, t14


@njit(fastmath=True)
def q1n(z, p, c, alpha, g, I_0):
    s = 1-z**2
    c0 = (1-c+c*s**g)
    c2 = 0.5*alpha*c*s**(g-2)*((alpha-1)*z**2-1)
    return 1-I_0*pi*p**2*(c0 + 0.25*p**2*c2 - 0.125*alpha*c*p**2*s**(g-1))


@njit(fastmath=True)
def q2n(z, p, c, alpha, g, I_0):
    d = (z**2 - p**2 + 1)/(2*z)
    ra = 0.5*(z-p+d)
    rb = 0.5*(1+d)
    sa = 1-ra**2
    sb = 1-rb**2
    q = (z-d)/p
    w2 = p**2-(d-z)**2
    w = sqrt(w2)
    b0 = 1 - c + c*sa**g
    b1 = -alpha*c*ra*sa**(g-1)
    b2 = 0.5*alpha*c*sa**(g-2)*((alpha-1)*ra**2-1)
    a0 = b0 + b1*(z-ra) + b2*(z-ra)**2
    a1 = b1+2*b2*(z-ra)
    aq = arccos(q)
    J1 = (a0*(d-z)-(2/3)*a1*w2 + 0.25*b2*(d-z)*(2*(d-z)**2-p**2))*w + (a0*p**2 + 0.25*b2*p**4)*aq
    J2 = alpha*c*sa**(g-1)*p**4*(0.125*aq + (1/12)*q*(q**2-2.5)*sqrt(1.-q**2) )
    d0 = 1 - c + c*sb**g
    d1 = -alpha*c*rb*sb**(g-1)
    K1 = ((d0-rb*d1)*arccos(d) + ((rb*d+(2/3)*(1-d**2))*d1 - d*d0)*sqrt(1-d**2))
    K2 = (1/3)*c*alpha*sb**(g+0.5)*(1-d)
    return 1. - I_0*(J1 - J2 + K1 - K2)


@njit(fastmath=False)
def qpower2_z_s(z, k, u):
    if k > 1.0 or any(u < 0.0):
        return nan
    if (copysign(1.0, z) < 0.0) or z >= 1.0 + k:
        flux = 1.0
    else:
        I_0 = (u[1] + 2) / (pi * (u[1] - u[0] * u[1] + 2))
        g = 0.5 * u[1]
        if z <= 1.0 - k:
            flux = q1n(z, k, u[0], u[1], g, I_0)
        elif abs(z - 1.0) < k:
            flux = q2n(z, k, u[0], u[1], g, I_0)
    return max(flux, 0.0)


@njit(parallel=True, fastmath=True)
def qpower2_model_v(t, k, ldc, t0, p, a, i, e, w, lcids, pbids, nsamples, exptimes):
    t0, p, a, i, e, w = atleast_1d(t0), atleast_1d(p), atleast_1d(a), atleast_1d(i), atleast_1d(e), atleast_1d(w)
    k = atleast_2d(k)
    ldc = atleast_2d(ldc)
    npv = k.shape[0]
    npt = t.size
    flux = zeros((npv, npt))
    for ipv in prange(npv):
        y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p[ipv], a[ipv], i[ipv], e[ipv], w[ipv])
        half_window_width = 0.025 + 0.5 * t14(k[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

        for j in range(npt):
            epoch = floor((t[j] - t0[ipv] + 0.5 * p[ipv]) / p[ipv])
            tc = t[j] - (t0[ipv] + epoch * p[ipv])
            if abs(tc) > half_window_width:
                flux[ipv, j] = 1.0
            else:
                ilc = lcids[j]
                ipb = pbids[ilc]

                if k.shape[1] == 1:
                    _k = k[ipv, 0]
                else:
                    _k = k[ipv, ipb]

                for isample in range(1, nsamples[ilc] + 1):
                    time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                    z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                    if z > 1.0 + _k:
                        flux[ipv, j] += 1.
                    else:
                        flux[ipv, j] += qpower2_z_s(z, _k, ldc[ipv, 2*ipb:2*(ipb+1)])
                flux[ipv, j] /= nsamples[ilc]
    return flux


@njit(parallel=False, fastmath=True)
def qpower2_model_s(t, k, ldc, t0, p, a, i, e, w, lcids, pbids, nsamples, exptimes):
    k = atleast_1d(k)
    ldc = atleast_1d(ldc)
    npt = t.size

    y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, i, e, w)
    half_window_width = 0.025 + 0.5 * t14(k[0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

    flux = zeros(npt)
    for j in range(npt):
        epoch = floor((t[j] - t0 + 0.5 * p) / p)
        tc = t[j] - (t0 + epoch * p)
        if abs(tc) > half_window_width:
            flux[j] = 1.0
        else:
            ilc = lcids[j]
            ipb = pbids[ilc]
            _k = k[0] if k.size == 1 else k[ipb]

            for isample in range(1, nsamples[ilc] + 1):
                time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                if z > 1.0 + _k:
                    flux[j] += 1.
                else:
                    flux[j] += qpower2_z_s(z, _k, ldc[2*ipb:2*(ipb+1)])
            flux[j] /= nsamples[ilc]
    return flux


@njit(parallel=True, fastmath=True)
def qpower2_model_pv(t, pvp, ldc, lcids, pbids, nsamples, exptimes):
    pvp = atleast_2d(pvp)
    ldc = atleast_2d(ldc)
    npv = pvp.shape[0]
    nk = pvp.shape[1] - 6
    npt = t.size
    flux = zeros((npv, npt))

    for ipv in prange(npv):
        t0, p, a, i, e, w = pvp[ipv, nk:]
        y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, i, e, w)
        half_window_width = 0.025 + 0.5 * t14(pvp[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

        for j in range(npt):
            epoch = floor((t[j] - t0 + 0.5*p)/p)
            tc = t[j] - (t0 + epoch*p)
            if abs(tc) > half_window_width:
                flux[ipv, j] = 1.0
            else:
                ilc = lcids[j]
                ipb = pbids[ilc]

                if nk == 1:
                    k = pvp[ipv, 0]
                else:
                    if ipb < nk:
                        k = pvp[ipv, ipb]
                    else:
                        k = nan

                for isample in range(1,nsamples[ilc]+1):
                    time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                    z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                    if z < 0.0 or z > 1.0+k:
                        flux[ipv, j] += 1.
                    else:
                        flux[ipv, j] += qpower2_z_s(z, k, ldc[ipv, 2*ipb:2*(ipb+1)])
                flux[ipv, j] /= nsamples[ilc]
    return flux



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to model flux contamination in transit light curves.
"""

from numpy import sqrt
from .instrument import Instrument
from .contamination import SMContamination, BBContamination, contaminate_light_curve
from .filter import DeltaFilter, BoxcarFilter, TabulatedFilter, sdss_z, sdss_i, sdss_r, sdss_g


def true_radius_ratio(apparent_k: float, contamination: float) -> float:
    return apparent_k / sqrt(1 - contamination)


def apparent_radius_ratio(true_k: float, contamination: float) -> float:
    return true_k * sqrt(1 - contamination)

sdss_griz = (sdss_g, sdss_r, sdss_i, sdss_z)

__all__ = ("SMContamination BBContamination Instrument DeltaFilter BoxcarFilter TabulatedFilter".split() +
           "true_radius_ratio apparent_radius_ratio sdss_z sdss_i sdss_r sdss_g sdss_griz".split() +
           ["contaminate_light_curve"])



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, Iterable

from astropy.constants import R_sun, M_sun
from matplotlib.patches import Circle
from matplotlib.pyplot import subplots, setp
from numpy import linspace, sin, cos, array, ndarray, asarray, squeeze, cross, newaxis, pi, where, nan, full, degrees, \
    zeros, polyfit, atleast_2d, atleast_1d, nansum, zeros_like
from numpy.linalg import norm
from scipy.interpolate import interp1d
from scipy.spatial.transform.rotation import Rotation

from .transitmodel import TransitModel
from .numba.gdmodel import create_star_xy, create_planet_xy, map_osm, xy_taylor_vt, oblate_model_s, \
    luminosity_v2, planck, create_star_luminosity
from ..contamination.filter import Filter, DeltaFilter
from ..stars import read_bt_settl_table, read_husser2013_table
from ..orbits import as_from_rhop, i_from_baew
from ..orbits.taylor_z import vajs_from_paiew, find_contact_point
from ..utils.octasphere import octasphere


def bt_settl_table(filters, tmin: float = 1200., tmax: float = 7000., nt: int = 200):
    """Calculates an interpolation table for passband-integrated flux as a function of temperature.
    """
    if tmax > 7000:
        raise ValueError('tmax for BT-SETTL model has to be smaller or equal to 7000 K.')
    if tmin < 1200:
        raise ValueError('tmax for BT-SETTL model has to be larger or equal to 1200 K.')

    temperatures = linspace(tmin, tmax, nt)
    npb = len(filters)

    spectra = read_bt_settl_table()
    wl_table = spectra.columns.values.astype('d')
    t_table = spectra.index.values
    spectra = interp1d(wl_table, spectra, bounds_error=False, fill_value=0.0)
    fluxes = zeros((npb, t_table.size))
    for i, f in enumerate(filters):
        wl, tr = f.sample()
        if wl.ndim == 0:
            fluxes[i] = spectra(wl) * tr
        else:
            fluxes[i] = (spectra(wl) * tr).mean(1)
    fluxes = interp1d(t_table, fluxes)(temperatures)
    fluxes /= fluxes.max(1)[:, newaxis]

    return fluxes, tmin, temperatures[1] - temperatures[0]


def husser2013_table(filters, tmin: float = 2500, tmax: float = 12000, nt: int = 200):
    """Calculates an interpolation table for passband-integrated flux as a function of temperature.
    """
    if tmax > 12000:
        raise ValueError('tmax for the Husser2013 model has to be smaller or equal to 12000 K.')
    if tmin < 2500:
        raise ValueError('tmax for the Husser2013 model has to be larger or equal to 2500 K.')

    temperatures = linspace(tmin, tmax, nt)
    npb = len(filters)

    spectra = read_husser2013_table()
    wl_table = spectra.columns.values.astype('d')
    t_table = spectra.index.values
    spectra = interp1d(wl_table, spectra, bounds_error=False, fill_value=0.0)
    fluxes = zeros((npb, t_table.size))
    for i, f in enumerate(filters):
        wl, tr = f.sample()
        if wl.ndim == 0:
            fluxes[i] = spectra(wl) * tr
        else:
            fluxes[i] = (spectra(wl) * tr).mean(1)
    fluxes = interp1d(t_table, fluxes)(temperatures)
    fluxes /= fluxes.max(1)[:, newaxis]
    return fluxes, tmin, temperatures[1] - temperatures[0]


def blackbody_table(filters, tmin: float = 1200, tmax: float = 12000, nt: int = 200):
    """Calculates an interpolation table for passband-integrated flux as a function of temperature.
    """
    temperatures = linspace(tmin, tmax, nt)
    npb = len(filters)

    fluxes = zeros((npb, nt))
    for i, f in enumerate(filters):
        wl, tr = f.sample()
        wl *= 1e-9
        for j, t in enumerate(temperatures):
            fluxes[i, j] = (planck(wl, t) * tr).mean()
    fluxes /= fluxes.max(1)[:,newaxis]
    return fluxes, tmin, temperatures[1] - temperatures[0]


class GravityDarkenedModel(TransitModel):
    """Transit model for a gravity-darkened fast-rotating oblate star.

    Transit model for a gravity-darkened fast-rotating oblate star following Barnes (ApJ, 2009, 705).
    """

    def __init__(self, filters: Union[float, Filter, Iterable[Filter]], rstar: float = 1.0,
                 sres: int = 80, pres: int = 5, tres: int = 60, model: str = 'blackbody',
                 tmin: float = 5000, tmax: float = 7000, accurate_mu: bool = True):
        """

        Parameters
        ----------
        filters
            Either the effective wavelength [nm], a Filter object, or a list of Filter objects.
        rstar
            Stellar radius [R_Sun]
        sres
            Stellar discretization resolution.
        pres
            Planet discretization resolution.
        tres
            Orbit discretization resolution.
        model
            The spectroscopic model to use. Can be either 'blackbody', 'husser2013', or 'bt-settl'.
        tmin
            Minimum allowed temperature [K].
        tmax
            Maximum allowed temperature [K].
        """
        super().__init__()

        assert model in ('blackbody', 'husser2013', 'bt-settl')
        self.model = model
        self.accurate_mu = accurate_mu

        if isinstance(filters, (float, int)):
            self.filters = (DeltaFilter('default', float(filters)),)
        elif isinstance(filters, Filter):
            self.filters = (filters,)
        else:
            self.filters = filters

        self.rstar = rstar*R_sun.value     # Stellar equator radius  [m]
        self.sres = sres                   # Integration resolution for the star
        self.pres = pres                   # Integration resolution for the planet
        self.tres = tres
        self.tmin = tmin                   # Minimum temperature [K]
        self.tmax = tmax                   # Maximum temperature [K]

        if self.model == 'blackbody':
            self._flux_table, self._teff0, self._dteff = blackbody_table(self.filters, tmin, tmax)
        elif self.model == 'husser2013':
            self._flux_table, self._teff0, self._dteff = husser2013_table(self.filters, tmin, tmax)
        else:
            self._flux_table, self._teff0, self._dteff = bt_settl_table(self.filters, tmin, tmax)

        self._ts, self._xs, self._ys = create_star_xy(sres)
        self._xp, self._yp = create_planet_xy(pres)

    def visualize(self, k, p, rho, b, e, w, alpha, rperiod, tpole, istar, beta, ldc, figsize=(5, 5), ax=None,
                  ntheta=18, vmin: float = 0.0, vmax: float = 1.0, passband: int = 0):
        if ax is None:
            fig, ax = subplots(figsize=figsize)
            ax.set_aspect(1.)
        else:
            fig, ax = None, ax

        ldc = atleast_2d(ldc)
        a = as_from_rhop(rho, p)
        inc = i_from_baew(b, a, e, w)
        mstar, ostar, gpole, f, _ = map_osm(rstar=self.rstar, rho=rho, rperiod=rperiod, tpole=tpole, phi=0.0)

        # Plot the star
        # -------------
        vertices_original, faces = octasphere(4)
        vertices = vertices_original.copy()
        vertices[:, 1] *= (1.0 - f)

        triangles = vertices[faces]
        centers = triangles.mean(1)
        normals = cross(triangles[:, 1] - triangles[:, 0], triangles[:, 2] - triangles[:, 0])
        nlength = norm(normals, axis=1)
        normals /= nlength[:, newaxis]

        rotation = Rotation.from_rotvec((0.5 * pi - istar) * array([1, 0, 0]))
        rn = rotation.apply(normals)
        rc = rotation.apply(centers)

        mask = rn[:, 2] < 0.0
        l = luminosity_v2(centers[mask], normals[mask], istar, mstar, self.rstar, ostar, tpole, gpole, beta,
                         ldc, passband, self._flux_table, self._teff0, self._dteff)
        l /= l.max()
        ax.tripcolor(rc[mask, 0], rc[mask, 1], l, shading='gouraud', vmin=vmin, vmax=vmax)

        nphi = 180
        theta = linspace(0 + 0.1, pi - 0.1, ntheta)
        phi = linspace(0, 2 * pi, nphi)
        for i in range(theta.size):
            y = (1.0 - f) * cos(theta[i])
            x = cos(phi) * sin(theta[i])
            z = sin(phi) * sin(theta[i])
            v = rotation.apply(array([x, full(nphi, y), z]).T)
            m = v[:, 2] < 0.0
            ax.plot(where(m, v[:, 0], nan), v[:, 1], 'k--', lw=1.5, alpha=0.25)

        # Plot the orbit
        # --------------
        y0, vx, vy, ax_, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, inc, e, w)
        c1 = find_contact_point(k, 1, y0, vx, vy, ax_, ay, jx, jy, sx, sy)
        c4 = find_contact_point(k, 4, y0, vx, vy, ax_, ay, jx, jy, sx, sy)
        time = linspace(2 * c1, 2 * c4, 100)

        ox, oy = xy_taylor_vt(time, alpha, y0, vx, vy, ax_, ay, jx, jy, sx, sy)
        ax.plot(ox, oy, 'k')

        pxy = xy_taylor_vt(array([0.0]), alpha, y0, vx, vy, ax_, ay, jx, jy, sx, sy)
        ax.add_artist(Circle(pxy, k, zorder=10, fc='k'))

        # Plot the info
        # -------------
        ax.text(0.025, 0.95, f"i$_\star$ = {degrees(istar):.1f}$^\circ$", transform=ax.transAxes)
        ax.text(0.025, 0.90, f"i$_\mathrm{{p}}$ = {degrees(inc):.1f}$^\circ$", transform=ax.transAxes)
        ax.text(1 - 0.025, 0.95, fr"$\alpha$ = {degrees(alpha):.1f}$^\circ$", transform=ax.transAxes, ha='right')
        ax.text(0.025, 0.05, f"f = {f:.1f}", transform=ax.transAxes)

        setp(ax, xlim=(-1.1, 1.1), ylim=(-1.1, 1.1), xticks=[], yticks=[])
        if fig is not None:
            fig.tight_layout()
        return ax

    def evaluate_ps(self, k: Union[float, ndarray], rho: float, rperiod: float, tpole: float, phi: float,
                    beta: float, ldc: ndarray, t0: float, p: float, a: float, i: float, l: float = 0.0,
                    e: float = 0.0, w: float = 0.0, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar parameters.

        Parameters
        ----------
        k : array-like
            Radius ratio(s) either as a single float or an 1D array
        rho : float
            Stellar density [g/cm^3]
        rperiod : float
            Stellar rotation period [d]
        tpole : float
            Temperature at the pole [K]
        phi : float
            Star's obliquity to the plane of the sky [rad]
        beta: float
            Gravity darkening parameter
        ldc : array-like
            Limb darkening coefficients as a 1D array
        t0 : float
            Zero epoch
        p : float
            Orbital period [d]
        a : float
            Scaled orbital semi-major axis [R_star]
        i : float
            Orbital inclination [rad]
        l : float
            Orbital azimuth angle [rad]
        e : float, optional
            Orbital eccentricity
        w : float, optional
            Argument of periastron

        Notes
        -----
        This version of the `evaluate` method is optimized for calculating a single transit model. If you want to
        evaluate the model for a large number of parameters simultaneously, use `evaluate`.

        Returns
        -------
        ndarray
            Modelled flux as a 1D ndarray.
        """

        ldc = asarray(ldc)
        k = asarray(k)

        if self.time is None:
            raise ValueError("Need to set the data before calling the transit model.")
        if ldc.size != 2*self.npb:
            raise ValueError("The quadratic model needs two limb darkening coefficients per passband")

        mstar, ostar, gpole, f, feff = map_osm(self.rstar, rho, rperiod, tpole, phi)
        sphi, cphi = sin(phi), cos(phi)

        flux = oblate_model_s(self.time, k, t0, p, a, l, i, e, w, ldc, mstar, self.rstar, ostar, tpole, gpole, f, feff,
                              sphi, cphi, beta, self._flux_table, self._teff0, self._dteff,
                              self.tres, self._ts, self._xs, self._ys,
                              self._xp, self._yp, self.lcids, self.pbids, self.nsamples, self.exptimes, self.npb, self.accurate_mu)

        return squeeze(flux)


    def evaluate_brute(self, k: Union[float, ndarray], rho: float, rperiod: float, tpole: float, phi: float,
                       beta: float, ldc: ndarray, t0: float, p: float, a: float, i: float, l: float = 0.0,
                       e: float = 0.0, w: float = 0.0, copy: bool = True, plot: bool = False, res: int = 300) -> ndarray:
        mstar, ostar, gpole, f, feff = map_osm(self.rstar, rho, rperiod, tpole, phi)
        sphi, cphi = sin(phi), cos(phi)
        ldc = atleast_2d(ldc)
        k = atleast_1d(k)

        st, sx, sy = create_star_xy(res)
        fstar = create_star_luminosity(res, sx, sy, mstar, self.rstar, ostar, tpole, gpole, f,
                                       sphi, cphi, beta, ldc, self._flux_table, self._teff0, self._dteff, self.accurate_mu)
        px, py = xy_taylor_vt(self.time - t0, l, *vajs_from_paiew(p, a, i, e, w))

        if plot:
            fig, ax = subplots()
            ax.imshow(fstar[0], extent=(-1, 1, -1, 1))
            ax.plot(px, py)
            fig.tight_layout()

        fnorm = [nansum(fstar[ipb]) for ipb in range(self.npb)]
        rflux = zeros_like(px)
        for j in range(px.size):
            ilc = self.lcids[j]
            ipb = self.pbids[ilc]
            m = ~((sx - px[j]) ** 2 + (sy - py[j]) ** 2 < k[0] ** 2).reshape((res, res))
            rflux[j] = nansum(fstar[ipb] * m) / fnorm[ipb]
        return rflux


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import astropy.units as u

from astropy.units import Quantity
from astropy.constants import G
from numpy import pi, sin, sqrt


def mp_from_kiepms(k, i, e, p, mstar):
    """Calculates the planet's mass from RV parameters.

    Parameters
    ----------
    k: float, ndarray, or Quantity
        RV semiamplitude [m/s]
    i: float, ndarray, or Quantity
        Inclination [rad]
    e: float, ndarray, or Quantity
        Eccentricity [-]
    p: float, ndarray, or Quantity
        Orbital period [d]
    mstar: float, ndarray, or Quantity
        Stellar mass [Solar mass]

    Returns
    -------
    Planet mass [M_Jup]
    """
    k = k.to(u.m/u.s) if isinstance(k, Quantity) else k * u.m/u.s
    i = i.to(u.rad) if isinstance(i, Quantity) else i*u.rad
    p = p.to(u.s) if isinstance(p, Quantity) else (p*u.day).to(u.s)
    mstar = mstar.to(u.kg) if isinstance(mstar, Quantity) else (mstar*u.M_sun).to(u.kg)
    return (k * (p/(2*pi*G)) ** (1/3) * mstar ** (2 / 3) / sin(i) * sqrt((1 - e ** 2))).to(u.M_jup).value


def surface_gravity(k, i, e, p, ar):
    """Calculates the planet's surface gravity from RV parameters.

    Parameters
    ----------
    k: float, ndarray, or Quantity
        RV semiamplitude [m/s]
    i: float, ndarray, or Quantity
        Inclination [rad]
    e: float, ndarray, or Quantity
        Eccentricity [-]
    p: float, ndarray, or Quantity
        Orbital period [d]
    ar: float, ndarray, or Quantity
        Planet-star area ratio [-]

    Returns
    -------
    Surface gravity [m/s^2]
    """
    k = k.to(u.m / u.s) if isinstance(k, Quantity) else k * u.m / u.s
    i = i.to(u.rad) if isinstance(i, Quantity) else i * u.rad
    p = p.to(u.s) if isinstance(p, Quantity) else (p * u.day).to(u.s)
    return (2 * pi / p * sqrt(1 - e ** 2) * k / (ar * sin(i))).value



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib.pyplot import subplots, setp
from numpy import pi, sign, cos, sqrt, sin, array, arccos, inf, round, int, s_, percentile, concatenate, median, mean, \
    arange

import scipy.ndimage as ndi

from numba import njit, prange
from .ttvlpf import TTVLPF
from ..models.transitmodel import TransitModel
from ..param.parameter import ParameterSet, PParameter, GParameter
from ..param.parameter import UniformPrior as U, NormalPrior as N, GammaPrior as GM
from ..orbits.orbits_py import duration_eccentric, as_from_rhop, p_from_dkaiews

try:
    import seaborn as sb
    with_seaborn = True
except ImportError:
    with_seaborn = False

@njit("f8[:](f8[:], f8[:], f8, f8, f8[:], i8[:])", cache=False, parallel=False)
def z_circular_ttv(t, p, a, i, tc, tcid):
    cosph = cos(2*pi * (t - tc[tcid]) / p[tcid])
    z = sign(cosph) * a * sqrt(1.0 - cosph * cosph * sin(i) ** 2)
    return z


class TDVLPF(TTVLPF):
    """Log posterior function for TDV estimation.

    A log posterior function for TDV estimation. Each light curve represents a single transit, and
    is given a separate free transit centre parameter. The average orbital period and (one) transit
    zero epoch are assumed as known.

    Notes: The number of parameters can grow large with Kepler short-period planets.

    """
    def __init__(self, target: str, zero_epoch: float, period: float, tc_sigma: float, duration_prior: tuple,
                 passbands: list, times: list = None, fluxes: list = None, pbids: list = None, tm: TransitModel = None,
                 nsamples: int = 1, exptime: float = 0.020433598):
        self.zero_epoch = zero_epoch
        self.period = period
        self.tc_sigma = tc_sigma
        self.t14_prior = duration_prior
        super().__init__(target, zero_epoch, period, tc_sigma, passbands, times, fluxes, pbids, tm, nsamples, exptime)

    def _init_p_orbit(self):
        """Orbit parameter initialisation for a TTV model.
           """

        # Basic orbital parameters
        # ------------------------
        porbit = [GParameter('rho', 'stellar_density', 'g/cm^3', U(0.1, 25.0), (0, inf)),
                  GParameter('b', 'impact_parameter', 'R_s', U(0.0, 1.0), (0, 1))]

        # Transit centers
        # ---------------

        def create_tc_prior(t, f, p=5):
            m = f > percentile(f, p)
            m = ~ndi.binary_erosion(m, iterations=6, border_value=1)
            return N(t[m].mean(), 0.25 * t[m].ptp())

        self.tnumber = round((array([t.mean() for t in self.times]) - self.zero_epoch) / self.period).astype(int)
        for t, f, tn in zip(self.times, self.fluxes, self.tnumber):
            prior = create_tc_prior(t, f)
            porbit.append(GParameter(f'tc_{tn:d}', f'transit_centre_{tn:d}', 'd', prior, (-inf, inf)))

        # Transit durations
        # -----------------
        for tn in self.tnumber:
            porbit.append(GParameter(f't14_{tn:d}', f'duration_{tn:d}', 'd', N(*self.t14_prior), (0, inf)))

        self.ps.add_global_block('orbit', porbit)
        self._start_tc = 2
        self._sl_tc = s_[self._start_tc:self._start_tc + self.nlc]
        self._start_d = 2 + self.nlc
        self._sl_d = s_[self._start_d:self._start_d + self.nlc]


    def _compute_z(self, pv):
        a = as_from_rhop(pv[0], self.period)
        if a < 1.:
            return None
        else:
            i = arccos(pv[1] / a)
            tc = pv[self._sl_tc]
            p = p_from_dkaiews(pv[self._sl_d], sqrt(pv[self._pid_k2]), a, i, 0., 0., 1)
            return z_circular_ttv(self.timea, p, a, i, tc, self.lcida)

    def plot_tdvs(self, burn=0, thin=1, ax=None, figsize=None, bwidth=0.8):
        fig, ax = (None, ax) if ax is not None else subplots(figsize=figsize)
        df = self.posterior_samples(burn, thin)
        dcols = [c for c in df.columns if 't14_' in c]
        p = 24 * df[dcols].quantile([0.50, 0.16, 0.84, 0.005, 0.995]).values
        ax.bar(self.tnumber, p[4, :] - p[3, :], bwidth, p[3, :], alpha=0.25, fc='b')
        ax.bar(self.tnumber, p[2, :] - p[1, :], bwidth, p[1, :], alpha=0.25, fc='b')
        [ax.plot((xx - 0.47 * bwidth, xx + 0.47 * bwidth), (pp[[0, 0]]), 'k') for xx, pp in zip(self.tnumber, p.T)]
        setp(ax, ylabel='Transit duration [h]', xlabel='Transit number')
        fig.tight_layout()
        if with_seaborn:
            sb.despine(ax=ax, offset=15)
        return ax


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit, prange
from numpy import pi, sqrt, arccos, abs, zeros_like, sign, sin, cos, abs, atleast_2d, zeros, atleast_1d, isnan, inf, \
    nan, copysign, fmax, floor
from ...orbits.taylor_z import vajs_from_paiew, z_taylor_st, vajs_from_paiew_eclipse, t14

TWO_PI = 2.0 * pi
HALF_PI = 0.5 * pi
FOUR_PI = 4.0 * pi
INV_PI = 1 / pi


@njit(cache=False, fastmath=True)
def uniform_z_v(zs, k, zsign=1.0):
    flux = zeros_like(zs)

    for i in range(len(zs)):
        z = zs[i] * zsign

        # Out of transit
        # --------------
        if (copysign(1.0, z) < 0.0) or z >= 1.0 + k:
            flux[i] = 1.0

        # Full transit by a larger object
        # -------------------------------
        elif k > 1.0 and z < k - 1.0:
            flux[i] = 0.0

        # Full transit
        # ------------
        elif z <= 1.0 - k:
            flux[i] = 1.0 - k * k

        # Ingress and egress
        # ------------------
        else:
            kap1 = arccos(min((1.0 - k * k + z * z) / 2.0 / z, 1.0))
            kap0 = arccos(min((k * k + z * z - 1.0) / 2.0 / k / z, 1.0))
            lambdae = k * k * kap0 + kap1
            lambdae = (lambdae - 0.5 * sqrt(max(4.0 * z * z - (1.0 + z * z - k * k) ** 2, 0.0))) / pi
            flux[i] = 1.0 - lambdae

    return flux


@njit(fastmath=True)
def uniform_z_s(z, k, zsign):
    z *= zsign

    # Out of transit
    # --------------
    if (copysign(1.0, z) < 0.0) or z >= 1.0 + k:
        flux = 1.0

    # Full transit by a larger object
    # -------------------------------
    elif k > 1.0 and z < k - 1.0:
        flux = 0.0

    # Full transit
    # ------------
    elif z <= 1.0 - k:
        flux = 1.0 - k * k

    # Ingress and egress
    # ------------------
    else:
        kap1 = arccos(min((1.0 - k * k + z * z) / 2.0 / z, 1.0))
        kap0 = arccos(min((k * k + z * z - 1.0) / 2.0 / k / z, 1.0))
        lambdae = k * k * kap0 + kap1
        lambdae = (lambdae - 0.5 * sqrt(max(4.0 * z * z - (1.0 + z * z - k * k) ** 2, 0.0))) / pi
        flux = 1.0 - lambdae

    return flux


@njit(parallel=True, fastmath=True)
def uniform_model_v(t, k, t0, p, a, i, e, w, lcids, pbids, nsamples, exptimes, zsign):
    t0, p, a, i, e, w = atleast_1d(t0), atleast_1d(p), atleast_1d(a), atleast_1d(i), atleast_1d(e), atleast_1d(w)
    k = atleast_2d(k)

    npv = k.shape[0]
    npt = t.size
    flux = zeros((npv, npt))
    for ipv in prange(npv):
        if zsign >= 0:
            y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p[ipv], a[ipv], i[ipv], e[ipv], w[ipv])
            et = 0.0
            half_window_width = 0.025 + 0.5 * t14(k[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)
        else:
            et, y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew_eclipse(p[ipv], a[ipv], i[ipv], e[ipv], w[ipv])
            half_window_width = 0.025 - 0.5 * t14(k[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

        for j in range(npt):
            epoch = floor((t[j] - t0[ipv] - et + 0.5 * p[ipv]) / p[ipv])
            tc = t[j] - (t0[ipv] + et + epoch * p[ipv])
            if abs(tc) > half_window_width:
                flux[ipv, j] = 1.0
            else:
                ilc = lcids[j]
                ipb = pbids[ilc]

                if a[ipv] < 1.0 or e[ipv] > 0.94:
                    flux[ipv, j] = nan
                    continue

                if k.shape[1] == 1:
                    _k = k[ipv, 0]
                else:
                    _k = k[ipv, ipb]

                for isample in range(1, nsamples[ilc] + 1):
                    time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                    z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                    if z > 1.0 + _k:
                        flux[ipv, j] += 1.
                    else:
                        flux[ipv, j] += uniform_z_s(z, _k, 1.0)
                flux[ipv, j] /= nsamples[ilc]
    return flux


@njit(parallel=False, fastmath=True)
def uniform_model_s(t, k, t0, p, a, i, e, w, lcids, pbids, nsamples, exptimes, zsign):
    k = atleast_1d(k)
    npt = t.size
    flux = zeros(npt)

    if a < 1.0:
        flux[:] = nan
        return flux

    if zsign >= 0:
        y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, i, e, w)
        et = 0.0
        half_window_width = 0.025 + 0.5 * t14(k[0], y0, vx, vy, ax, ay, jx, jy, sx, sy)
    else:
        et, y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew_eclipse(p, a, i, e, w)
        half_window_width = 0.025 - 0.5 * t14(k[0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

    for j in range(npt):
        epoch = floor((t[j] - t0 - et + 0.5 * p) / p)
        tc = t[j] - (t0 + et + epoch * p)
        if abs(tc) > half_window_width:
            flux[j] = 1.0
        else:
            ilc = lcids[j]
            ipb = pbids[ilc]
            _k = k[0] if k.size == 1 else k[ipb]

            for isample in range(1, nsamples[ilc] + 1):
                time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                if z > 1.0 + _k:
                    flux[j] += 1.
                else:
                    flux[j] += uniform_z_s(z, _k, 1.0)
            flux[j] /= nsamples[ilc]
    return flux


@njit(parallel=True, fastmath=False)
def uniform_model_pv(t, pvp, lcids, pbids, nsamples, exptimes, zsign):
    pvp = atleast_2d(pvp)
    npv = pvp.shape[0]
    npt = t.size
    nk = pvp.shape[1] - 6

    flux = zeros((npv, npt))
    for ipv in range(npv):
        t0, p, a, i, e, w = pvp[ipv, nk:]

        if zsign >= 0:
            y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew(p, a, i, e, w)
            et = 0.0
            half_window_width = 0.025 + 0.5 * t14(pvp[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)
        else:
            et, y0, vx, vy, ax, ay, jx, jy, sx, sy = vajs_from_paiew_eclipse(p, a, i, e, w)
            half_window_width = 0.025 - 0.5 * t14(pvp[ipv, 0], y0, vx, vy, ax, ay, jx, jy, sx, sy)

        for j in prange(npt):
            epoch = floor((t[j] - t0 - et + 0.5 * p) / p)
            tc = t[j] - (t0 + et + epoch * p)
            if abs(tc) > half_window_width:
                flux[ipv, j] = 1.0
            else:
                ilc = lcids[j]
                ipb = pbids[ilc]

                if a < 1.0 or e > 0.94:
                    flux[ipv, j] = nan
                    continue

                if nk == 1:
                    k = pvp[ipv, 0]
                else:
                    if ipb < nk:
                        k = pvp[ipv, ipb]
                    else:
                        k = nan

                for isample in range(1,nsamples[ilc]+1):
                    time_offset = exptimes[ilc] * ((isample - 0.5) / nsamples[ilc] - 0.5)
                    z = z_taylor_st(tc + time_offset, y0, vx, vy, ax, ay, jx, jy, sx, sy)
                    if z > 1.0+k:
                        flux[ipv, j] += 1.
                    else:
                        flux[ipv, j] += uniform_z_s(z, k, 1.0)
                flux[ipv, j] /= nsamples[ilc]
    return flux



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
from typing import Union, Optional

import numpy as np
import pyopencl as cl
from os.path import dirname, join

import warnings

from pkg_resources import resource_filename
from pyopencl import CompilerWarning

from numpy import array, uint32, float32, int32, asarray, zeros, ones, unique, atleast_2d, squeeze, ndarray, empty, \
    concatenate

from .numba.ma_quadratic_nb import calculate_interpolation_tables
from .transitmodel import TransitModel

warnings.filterwarnings('ignore', category=CompilerWarning)

class QuadraticModelCL(TransitModel):
    """
    OpenCL implementation of the transit light curve model with quadratic limb darkening by Mandel and Agol (2002).

    This class implements the quadratic transit model by Mandel & Agol (ApJ 580, L171-L175, 2002) in OpenCL. The class
    can replace `pytransit.QuadraticModel` directly, and offers in most cases a significant performance boost with some
    drawbacks (see the notes below).

    Notes
    ----
    - All the calculations are done in **single precision**. This can affect the results when modelling extremely shallow
      transits, and will certainly affect the model if the times are given in JD. The times should be given relative to
      some constant epoch that maximises the precision (see `pytransit.lpf.lpf.BaseLPF` for an example).

    """

    def __init__(self, klims: tuple = (0.05, 0.25), nk: int = 256, nz: int = 256, cl_ctx=None, cl_queue=None) -> None:
        """Transit model with quadratic limb darkening (Mandel & Agol, ApJ 580, L171-L175, 2002).

        Parameters
        ----------
        klims : tuple, optional
            Radius ratio limits (kmin, kmax) for the interpolated model.
        nk : int, optional
            Radius ratio grid size for the interpolated model.
        nz : int, optional
            Normalized distance grid size for the interpolated model.
        cl_ctx: optional
            OpenCL context.
        cl_queue: optional
            OpenCL queue

        """
        super().__init__()

        self.ctx = cl_ctx or cl.create_some_context()
        self.queue = cl_queue or cl.CommandQueue(self.ctx)

        self.ed,self.le,self.ld,self.kt,self.zt = map(lambda a: np.array(a,dtype=float32,order='C'),
                                                      calculate_interpolation_tables(klims[0],klims[1],nk,nz))
        self.klims = klims
        self.nk    = int32(nk)
        self.nz    = int32(nz)
        self.nptb  = 0
        self.npb   = 0
        self.u     = np.array([])
        self.f     = None
        self.k0, self.k1 = map(float32, self.kt[[0,-1]])
        self.dk = float32(self.kt[1]-self.kt[0])
        self.dz = float32(self.zt[1]-self.zt[0])
        self.pv = array([])

        mf = cl.mem_flags

        # Create the buffers for the Mandel & Agol coefficient arrays
        self._b_ed = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.ed)
        self._b_le = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.le)
        self._b_ld = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.ld)
        self._b_kt = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.kt)
        self._b_zt = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.zt)

        self.time  = None
        self.lcids = None
        self.pbids = None
        self.nsamples = None
        self.exptimes = None
        self.vajs:     Optional[ndarray] = None

        # Declare the buffers for the ld coefficients, time, and flux arrays. These will
        # be initialised when the model is first evaluated, and reinitialised if the
        # array sizes change.
        #
        self._b_u = None       # Limb darkening coefficient buffer
        self._b_time = None    # Time buffer
        self._b_f = None       # Flux buffer
        self._b_p = None       # Parameter vector buffer
        self._b_vajs = None

        self._time_id = None   # Time array ID

        rd = Path(resource_filename('pytransit', 'models/opencl'))
        po = rd / 'orbits.cl'
        pm = rd / 'ma_quadratic.cl'
        self.prg = cl.Program(self.ctx, po.read_text() + pm.read_text()).build()

    def set_data(self, time, lcids=None, pbids=None, nsamples=None, exptimes=None):
        mf = cl.mem_flags

        if self._b_time is not None:
            self._b_time.release()
            self._b_lcids.release()
            self._b_pbids.release()
            self._b_nsamples.release()
            self._b_etimes.release()

        self.nlc = uint32(1 if lcids is None else unique(lcids).size)
        self.npb = uint32(1 if pbids is None else unique(pbids).size)
        self.nptb = time.size

        #TODO: Fix nsamples and exptimes in cases where a scalar is given.
        self.time = asarray(time, dtype='float32')
        self.lcids = zeros(time.size, 'uint32') if lcids is None else asarray(lcids, dtype='uint32')
        self.pbids = zeros(self.nlc, 'uint32') if pbids is None else asarray(pbids, dtype='uint32')
        self.nsamples = ones(self.nlc, 'uint32') if nsamples is None else asarray(nsamples, dtype='uint32')
        self.exptimes = ones(self.nlc, 'float32') if exptimes is None else asarray(exptimes, dtype='float32')

        self._b_time = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.time)
        self._b_lcids = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.lcids)
        self._b_pbids = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pbids)
        self._b_nsamples = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.nsamples)
        self._b_etimes = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.exptimes)

    def evaluate(self, k: Union[float, ndarray], ldc: ndarray, t0: Union[float, ndarray], p: Union[float, ndarray],
                 a: Union[float, ndarray], i: Union[float, ndarray], e: Union[float, ndarray] = None,
                 w: Union[float, ndarray] = None, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar or vector parameters.

        Parameters
        ----------
        k
            Radius ratio(s) either as a single float, 1D vector, or 2D array.
        ldc
            Limb darkening coefficients as a 1D or 2D array.
        t0
            Transit center(s) as a float or a 1D vector.
        p
            Orbital period(s) as a float or a 1D vector.
        a
            Orbital semi-major axis (axes) divided by the stellar radius as a float or a 1D vector.
        i
            Orbital inclination(s) as a float or a 1D vector.
        e : optional
            Orbital eccentricity as a float or a 1D vector.
        w : optional
            Argument of periastron as a float or a 1D vector.

        Notes
        -----
        The model can be evaluated either for one set of parameters or for many sets of parameters simultaneously. In
        the first case, the orbital parameters should all be given as floats. In the second case, the orbital parameters
        should be given as a 1D array-like.

        Returns
        -------
        ndarray
            Modelled flux either as a 1D or 2D ndarray.
        """
        npv = 1 if isinstance(t0, float) else len(t0)
        k = asarray(k)

        if k.size == 1:
            nk = 1
        elif npv == 1:
            nk = k.size
        else:
            nk = k.shape[1]

        if e is None:
            e, w = 0.0, 0.0

        pvp = empty((npv, nk + 6), dtype=float32)
        pvp[:, :nk] = k
        pvp[:, nk] = t0
        pvp[:, nk + 1] = p
        pvp[:, nk + 2] = a
        pvp[:, nk + 3] = i
        pvp[:, nk + 4] = e
        pvp[:, nk + 5] = w

        ldc = atleast_2d(ldc).astype(float32)
        self.npv = uint32(pvp.shape[0])
        self.spv = uint32(pvp.shape[1])

        # Release and reinitialise the GPU buffers if the sizes of the time or
        # limb darkening coefficient arrays change.
        if (ldc.size != self.u.size) or (pvp.size != self.pv.size):
            assert self.npb == ldc.shape[1] // 2

            if self._b_f is not None:
                self._b_f.release()
                self._b_u.release()
                self._b_p.release()
                self._b_vajs.release()

            self.pv = zeros(pvp.shape, float32)
            self.u = zeros((self.npv, 2 * self.npb), float32)
            self.f = zeros((self.npv, self.nptb), float32)
            self.vajs = zeros((self.npv, 9), float32)

            mf = cl.mem_flags
            self._b_f = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.time.nbytes * self.npv)
            self._b_u = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.u)
            self._b_p = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pv)
            self._b_vajs = cl.Buffer(self.ctx, mf.READ_WRITE, float32().nbytes*self.npv*9)

        # Copy the limb darkening coefficient array to the GPU
        cl.enqueue_copy(self.queue, self._b_u, ldc)

        # Copy the parameter vector to the GPU
        self.pv[:] = pvp
        cl.enqueue_copy(self.queue, self._b_p, self.pv)

        self.prg.vajs_from_paiew_v(self.queue, (self.npv, ), None, self.spv, self._b_p, self._b_vajs)
        self.prg.ma_eccentric_pop(self.queue, (self.npv, self.nptb), None, self._b_time, self._b_lcids, self._b_pbids,
                                  self._b_p, self._b_u,
                                  self._b_ed, self._b_le, self._b_ld, self._b_nsamples, self._b_etimes, self._b_vajs,
                                  self.k0, self.k1, self.nk, self.nz, self.dk, self.dz,
                                  self.spv, self.nlc, self.npb, self._b_f)

        if copy:
            cl.enqueue_copy(self.queue, self.f, self._b_f)
            return squeeze(self.f)
        else:
            return None

    def evaluate_ps(self, k: Union[float, ndarray], ldc: ndarray, t0: float, p: float, a: float, i: float,
                    e: float = 0.0, w: float = 0.0, copy: bool = True):
        """Evaluate the transit model for a set of scalar parameters.

        Parameters
        ----------
        k : array-like
            Radius ratio(s) either as a single float or an 1D array.
        ldc : array-like
            Limb darkening coefficients as a 1D array.
        t0 : float
            Transit center as a float.
        p : float
            Orbital period as a float.
        a : float
            Orbital semi-major axis divided by the stellar radius as a float.
        i : float
            Orbital inclination(s) as a float.
        e : float, optional
            Orbital eccentricity as a float.
        w : float, optional
            Argument of periastron as a float.

        Returns
        -------
        ndarray
            Modelled flux as a 1D ndarray.
        """
        if isinstance(k, float):
            pv = array([[k, t0, p, a, i, e, w]], float32)
        else:
            pv = concatenate([k, [t0, p, a, i, e, w]]).astype(float32)
        return self.evaluate_pv(pv, ldc, copy)

    def evaluate_pv(self, pvp: ndarray, ldc: ndarray, copy: bool = True):
        """Evaluate the transit model for 2D parameter array.

         Parameters
         ----------
         pvp
             Parameter array with a shape `(npv, npar)` where `npv` is the number of parameter vectors, and each row
             contains a set of parameters `[k, t0, p, a, i, e, w]`. The radius ratios can also be given per passband,
             in which case the row should be structured as `[k_0, k_1, k_2, ..., k_npb, t0, p, a, i, e, w]`.
         ldc
             Limb darkening coefficient array with shape `(npv, 2*npb)`, where `npv` is the number of parameter vectors
             and `npb` is the number of passbands.

         Notes
         -----
         This version of the `evaluate` method is optimized for calculating several models in parallel, such as when
         using *emcee* for MCMC sampling.

         Returns
         -------
         ndarray
             Modelled flux either as a 1D or 2D ndarray.
         """
        pvp = atleast_2d(pvp)
        ldc = atleast_2d(ldc).astype(float32)
        self.npv = uint32(pvp.shape[0])
        self.spv = uint32(pvp.shape[1])

        if pvp.shape[0] != ldc.shape[0]:
            raise ValueError("The parameter array and the ldc array have incompatible dimensions.")

        # Release and reinitialise the GPU buffers if the sizes of the time or
        # limb darkening coefficient arrays change.
        if (ldc.size != self.u.size) or (pvp.size != self.pv.size):
            assert self.npb == ldc.shape[1] // 2

            if self._b_f is not None:
                self._b_f.release()
                self._b_u.release()
                self._b_p.release()
                self._b_vajs.release()

            self.pv = zeros(pvp.shape, float32)
            self.u = zeros((self.npv, 2 * self.npb), float32)
            self.f = zeros((self.npv, self.nptb), float32)
            self.vajs = zeros((self.npv, 9), float32)

            mf = cl.mem_flags
            self._b_f = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.time.nbytes * self.npv)
            self._b_u = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.u)
            self._b_p = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pv)
            self._b_vajs = cl.Buffer(self.ctx, mf.READ_WRITE, float32().nbytes*self.npv*9)

        # Copy the limb darkening coefficient array to the GPU
        cl.enqueue_copy(self.queue, self._b_u, ldc)

        # Copy the parameter vector to the GPU
        self.pv[:] = pvp
        cl.enqueue_copy(self.queue, self._b_p, self.pv)

        self.prg.vajs_from_paiew_v(self.queue, (self.npv, ), None, self.spv, self._b_p, self._b_vajs)
        self.prg.ma_eccentric_pop(self.queue, (self.npv, self.nptb), None, self._b_time, self._b_lcids, self._b_pbids,
                                  self._b_p, self._b_u,
                                  self._b_ed, self._b_le, self._b_ld, self._b_nsamples, self._b_etimes, self._b_vajs,
                                  self.k0, self.k1, self.nk, self.nz, self.dk, self.dz,
                                  self.spv, self.nlc, self.npb, self._b_f)

        if copy:
            cl.enqueue_copy(self.queue, self.f, self._b_f)
            return squeeze(self.f)
        else:
            return None

    def evaluate_pv_ttv(self, pvp: ndarray, ldc: ndarray, copy: bool = True, tdv: bool = False):
        pvp = atleast_2d(pvp)
        ldc = atleast_2d(ldc).astype(float32)
        self.npv = uint32(pvp.shape[0])
        self.spv = uint32(pvp.shape[1])

        # Release and reinitialise the GPU buffers if the sizes of the time or
        # limb darkening coefficient arrays change.
        if (ldc.size != self.u.size) or (pvp.size != self.pv.size):
            assert self.npb == ldc.shape[1] // 2

            if self._b_f is not None:
                self._b_f.release()
                self._b_u.release()
                self._b_p.release()

            self.pv = zeros(pvp.shape, float32)
            self.u  = zeros((self.npv, 2*self.npb), float32)
            self.f  = zeros((self.npv, self.time.size), float32)

            mf = cl.mem_flags
            self._b_f   = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.time.nbytes * self.npv)
            self._b_u   = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.u)
            self._b_p   = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pv)

        # Copy the limb darkening coefficient array to the GPU
        cl.enqueue_copy(self.queue, self._b_u, ldc)

        # Copy the parameter vector to the GPU
        self.pv[:] = pvp
        cl.enqueue_copy(self.queue, self._b_p, self.pv)

        if tdv:
            self.prg.ma_eccentric_pop_tdv(self.queue, (self.npv, self.nptb), None, self._b_time, self._b_lcids,
                                          self._b_pbids,
                                          self._b_p, self._b_u,
                                          self._b_ed, self._b_le, self._b_ld, self._b_nsamples, self._b_etimes,
                                          self.k0, self.k1, self.nk, self.nz, self.dk, self.dz,
                                          self.spv, self.nlc, self.npb, self._b_f)
        else:
            self.prg.ma_eccentric_pop_ttv(self.queue, (self.npv, self.nptb), None,
                                          self._b_time, self._b_lcids, self._b_pbids,
                                          self._b_p, self._b_u,
                                          self._b_ed, self._b_le, self._b_ld,
                                          self._b_nsamples, self._b_etimes,
                                          self.k0, self.k1, self.nk, self.nz, self.dk, self.dz,
                                          self.spv, self.nlc, self.npb,
                                          self._b_f)

        if copy:
            cl.enqueue_copy(self.queue, self.f, self._b_f)
            return self.f
        else:
            return None



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import warnings

from numpy import exp
from scipy.constants import c, k, h
from numba import jit

warnings.warn("the pytransits.utils.physics module is deprecated and will be removed in PyTransit 2.0",
               FutureWarning, stacklevel=2)

@jit
def planck(T, wl):
    """Radiance as a function or black-body temperature and wavelength.

    Parameters
    ----------

      T   : Temperature  [K]
      wl  : Wavelength   [m]

    Returns
    -------

      B   : Radiance
    """
    return 2*h*c**2 / wl**5 / (exp(h*c / (wl*k*T)) - 1.)

@jit
def planck_ratio(T1, T2, wl):
    """Ratio of the two black-body object radiances

    Parameters
    ----------

      T1  : Temperature  [K]
      T2  : Temperature  [K]
      wl  : Wavelength   [m]

    Returns
    -------

      rB   : Radiance ratio
    """
    return  (exp(h*c / (wl*k*T2)) - 1.) / (exp(h*c / (wl*k*T1)) - 1.)